"""Fan Devices Base Class (NOT purifiers or humidifiers)."""

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import FanFeatures, FanModes
from pyvesync.utils.helpers import OscillationCoordinates, OscillationRange

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import FanMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class FanState(DeviceState):
    """Base state class for Fans.

    Not all attributes are supported by all devices.


    Attributes:
        display_set_status (str): Display set status.
        display_status (str): Display status.
        displaying_type (str): Displaying type.
        fan_level (int): Fan level.
        fan_set_level (int): Fan set level.
        humidity (int): Humidity level.
        mode (str): Mode of device.
        mute_set_status (str): Mute set status.
        mute_status (str): Mute status.
        oscillation_set_status (str): Oscillation set status.
        oscillation_status (str): Oscillation status.
        sleep_change_fan_level (str): Sleep change fan level.
        sleep_fallasleep_remain (str): Sleep fall asleep remain.
        sleep_oscillation_switch (str): Sleep oscillation switch.
        sleep_preference_type (str): Sleep preference type.
        temperature (int): Temperature.
        thermal_comfort (int): Thermal comfort.
        timer (Timer): Timer object.
    """

    __slots__ = (
        '_oscillation_status',
        'child_lock',
        'display_set_status',
        'display_status',
        'displaying_type',
        'fan_level',
        'fan_set_level',
        'horizontal_oscillation_status',
        'humidity',
        'mode',
        'mute_set_status',
        'mute_status',
        'oscillation_coordinates',
        'oscillation_range',
        'oscillation_set_status',
        'sleep_change_fan_level',
        'sleep_fallasleep_remain',
        'sleep_oscillation_switch',
        'sleep_preference_type',
        'temperature',
        'thermal_comfort',
        'vertical_oscillation_status',
    )

    def __init__(
        self,
        device: VeSyncFanBase,
        details: ResponseDeviceDetailsModel,
        feature_map: FanMap,
    ) -> None:
        """Initialize Fan State.

        Args:
            device (VeSyncFanBase): Device object.
            details (ResponseDeviceDetailsModel): Device details.
            feature_map (FanMap): Feature map.
        """
        super().__init__(device, details, feature_map)
        self.mode: str | None = None
        self.fan_level: int | None = None
        self.fan_set_level: int | None = None
        self.child_lock: bool | None = None
        self.humidity: float | None = None
        self.temperature: float | None = None
        self.thermal_comfort: int | None = None
        self.sleep_preference_type: str | None = None
        self.sleep_fallasleep_remain: str | None = None
        self.sleep_oscillation_switch: str | None = None
        self.sleep_change_fan_level: str | None = None
        self.mute_status: str | None = None
        self.mute_set_status: str | None = None
        self.horizontal_oscillation_status: str | None = None
        self.vertical_oscillation_status: str | None = None
        self._oscillation_status: str | None = None
        self.oscillation_set_status: str | None = None
        self.display_status: str | None = None
        self.display_set_status: str | None = None
        self.displaying_type: str | None = None
        self.oscillation_coordinates: OscillationCoordinates | None = None
        if FanFeatures.SET_OSCILLATION_RANGE in feature_map.features:
            self.oscillation_range: OscillationRange | None = OscillationRange(
                left=0, right=0, top=0, bottom=0
            )
        else:
            self.oscillation_range = None

    @property
    def oscillation_status(self) -> str | None:
        """Get oscillation status.

        Returns:
            str | None: Oscillation status.
        """
        if self.horizontal_oscillation_status and self.vertical_oscillation_status:
            return self.horizontal_oscillation_status or self.vertical_oscillation_status
        return self._oscillation_status

    @oscillation_status.setter
    def oscillation_status(self, value: str | None) -> None:
        self._oscillation_status = value


class VeSyncFanBase(VeSyncBaseToggleDevice):
    """Base device for VeSync tower fans.

    Inherits from
    [VeSyncBaseToggleDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseToggleDevice]
    and [VeSyncBaseDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseDevice].

    Attributes:
        fan_levels (list[int]): Fan levels supported by device.
        modes (list[str]): Modes supported by device.
        sleep_preferences (list[str]): Sleep preferences supported by device.
    """

    __slots__ = (
        '_reverse_modes',
        'fan_levels',
        'modes',
        'sleep_preferences',
    )

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: FanMap,
    ) -> None:
        """Initialize VeSync Tower Fan Base Class.

        Args:
            details (ResponseDeviceDetailsModel): Device details.
            manager (VeSync): Manager.
            feature_map (FanMap): Feature map.

        See Also:
            See [device_map][pyvesync.device_map] for configured features and modes.
        """
        super().__init__(details, manager, feature_map)
        self.features: list[str] = feature_map.features
        self.state: FanState = FanState(self, details, feature_map)
        self.modes: dict[str, str] = feature_map.modes
        self._reverse_modes: dict[str, str] = {}
        for k, v in self.modes.items():
            self._reverse_modes[v] = k
        self.fan_levels: list[int] = feature_map.fan_levels
        self.sleep_preferences: list[str] = feature_map.sleep_preferences

    @property
    def supports_oscillation(self) -> bool:
        """Return True if device supports oscillation."""
        return FanFeatures.OSCILLATION in self.features

    @property
    def supports_set_oscillation_range(self) -> bool:
        """Return True if device supports setting oscillation range."""
        return FanFeatures.SET_OSCILLATION_RANGE in self.features

    @property
    def supports_vertical_oscillation(self) -> bool:
        """Return True if device supports vertical oscillation."""
        return FanFeatures.VERTICAL_OSCILLATION in self.features

    @property
    def supports_horizontal_oscillation(self) -> bool:
        """Return True if device supports horizontal oscillation."""
        return FanFeatures.HORIZONTAL_OSCILLATION in self.features

    @property
    def supports_mute(self) -> bool:
        """Return True if device supports mute."""
        return FanFeatures.SOUND in self.features

    @property
    def supports_displaying_type(self) -> bool:
        """Return True if device supports displaying type."""
        return FanFeatures.DISPLAYING_TYPE in self.features

    async def toggle_display(self, toggle: bool) -> bool:
        """Toggle Display on/off.

        Args:
            toggle (bool): Display state.

        Returns:
            bool: Success of request.
        """
        del toggle
        return False

    async def turn_on_display(self) -> bool:
        """Turn on Display.

        Returns:
            bool: Success of request
        """
        return await self.toggle_display(True)

    async def turn_off_display(self) -> bool:
        """Turn off Display.

        Returns:
            bool: Success of request
        """
        return await self.toggle_display(False)

    @abstractmethod
    async def set_mode(self, mode: str) -> bool:
        """Set Fan Mode.

        Args:
            mode (str): Mode to set, varies by device type.

        Returns:
            bool: Success of request.
        """

    @abstractmethod
    async def set_fan_speed(self, speed: int | None = None) -> bool:
        """Set Fan Fan Speed.

        Args:
            speed (int, optional): Fan speed level according to device specs.

        Returns:
            bool: Success of request.
        """

    async def set_auto_mode(self) -> bool:
        """Set Fan to Auto Mode.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.AUTO in self.modes:
            return await self.set_mode(FanModes.AUTO)
        logger.warning('Auto mode not supported for this device.')
        return False

    async def set_advanced_sleep_mode(self) -> bool:
        """Set Fan to Advanced Sleep Mode.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.SLEEP in self.modes:
            return await self.set_mode(FanModes.SLEEP)
        logger.warning('Sleep mode not supported for this device.')
        return False

    async def set_sleep_mode(self) -> bool:
        """Set Fan to Sleep Mode.

        This is also referred to as Advanced Sleep Mode on some devices.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.SLEEP in self.modes:
            return await self.set_mode(FanModes.SLEEP)
        logger.warning('Sleep mode not supported for this device.')
        return False

    async def set_manual_mode(self) -> bool:
        """Set Fan to Manual Mode - Normal Mode.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.NORMAL in self.modes:
            return await self.set_mode(FanModes.NORMAL)
        logger.warning('Manual mode not supported for this device.')
        return False

    async def set_normal_mode(self) -> bool:
        """Set Fan to Normal Mode.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.NORMAL in self.modes:
            return await self.set_mode(FanModes.NORMAL)
        logger.warning('Normal mode not supported for this device.')
        return False

    async def set_turbo_mode(self) -> bool:
        """Set Fan to Turbo Mode.

        Returns:
            bool: Success of request.

        Note:
            This method is not supported by all devices, will return
            false with warning debug message if not supported.
        """
        if FanModes.TURBO in self.modes:
            return await self.set_mode(FanModes.TURBO)
        logger.warning('Turbo mode not supported for this device.')
        return False

    async def toggle_oscillation(self, toggle: bool) -> bool:
        """Toggle Oscillation on/off.

        Args:
            toggle (bool): Oscillation state.

        Returns:
            bool: true if success.
        """
        del toggle
        if self.supports_oscillation:
            logger.debug('Oscillation not configured for this device.')
        else:
            logger.debug('Oscillation not supported for this device.')
        return False

    async def turn_on_oscillation(self) -> bool:
        """Set toggle_oscillation to on."""
        return await self.toggle_oscillation(True)

    async def turn_off_oscillation(self) -> bool:
        """Set toggle_oscillation to off."""
        return await self.toggle_oscillation(False)

    async def toggle_vertical_oscillation(self, toggle: bool) -> bool:
        """Toggle Vertical Oscillation on/off.

        Args:
            toggle (bool): Vertical Oscillation state.

        Returns:
            bool: true if success.
        """
        del toggle
        if self.supports_vertical_oscillation:
            logger.debug('Vertical Oscillation not configured for this device.')
        else:
            logger.debug('Vertical Oscillation not supported for this device.')
        return False

    async def turn_on_vertical_oscillation(self) -> bool:
        """Set toggle_vertical_oscillation to on."""
        return await self.toggle_vertical_oscillation(True)

    async def turn_off_vertical_oscillation(self) -> bool:
        """Set toggle_vertical_oscillation to off."""
        return await self.toggle_vertical_oscillation(False)

    async def toggle_horizontal_oscillation(self, toggle: bool) -> bool:
        """Toggle Horizontal Oscillation on/off.

        Args:
            toggle (bool): Horizontal Oscillation state.

        Returns:
            bool: true if success.
        """
        del toggle
        if self.supports_horizontal_oscillation:
            logger.debug('Horizontal Oscillation not configured for this device.')
        else:
            logger.debug('Horizontal Oscillation not supported for this device.')
        return False

    async def turn_on_horizontal_oscillation(self) -> bool:
        """Set toggle_horizontal_oscillation to on."""
        return await self.toggle_horizontal_oscillation(True)

    async def turn_off_horizontal_oscillation(self) -> bool:
        """Set toggle_horizontal_oscillation to off."""
        return await self.toggle_horizontal_oscillation(False)

    async def set_vertical_oscillation_range(
        self, *, top: int = 0, bottom: int = 0
    ) -> bool:
        """Set Vertical Oscillation Range.

        Args:
            top (int): Top range.
            bottom (int): Bottom range.

        Returns:
            bool: true if success.
        """
        del top, bottom
        if self.supports_set_oscillation_range:
            logger.debug('Vertical oscillation range not configured for this device.')
        else:
            logger.debug('Vertical oscillation range not supported for this device.')
        return False

    async def set_horizontal_oscillation_range(
        self, *, left: int = 0, right: int = 0
    ) -> bool:
        """Set Horizontal Oscillation Range.

        Args:
            left (int): Left range.
            right (int): Right range.

        Returns:
            bool: true if success.
        """
        del left, right
        if self.supports_set_oscillation_range:
            logger.debug('Horizontal oscillation range not configured for this device.')
        else:
            logger.debug('Horizontal oscillation range not supported for this device.')
        return False

    async def toggle_mute(self, toggle: bool) -> bool:
        """Toggle mute on/off.

        Parameters:
            toggle (bool): True to turn mute on, False to turn off

        Returns:
            bool : True if successful, False if not
        """
        del toggle
        if self.supports_mute:
            logger.debug('Mute not configured for this device.')
        else:
            logger.debug('Mute not supported for this device.')
        return False

    async def turn_on_mute(self) -> bool:
        """Set toggle_mute to on."""
        return await self.toggle_mute(True)

    async def turn_off_mute(self) -> bool:
        """Set toggle_mute to off."""
        return await self.toggle_mute(False)

    async def toggle_displaying_type(self, toggle: bool) -> bool:
        """Toggle displaying type on/off.

        This functionality is unknown but was in the API calls.

        Args:
            toggle (bool): Displaying type state.

        Returns:
            bool: true if success.
        """
        del toggle
        if self.supports_displaying_type:
            logger.debug('Displaying type not configured for this device.')
        else:
            logger.debug('Displaying type not supported for this device.')
        return False

    @deprecated('Use `set_normal_mode` method instead')
    async def normal_mode(self) -> bool:
        """Set mode to normal."""
        return await self.set_normal_mode()

    @deprecated('Use `set_manual_mode` method instead')
    async def manual_mode(self) -> bool:
        """Adapter to set mode to normal."""
        return await self.set_normal_mode()

    @deprecated('Use `set_advanced_sleep_mode` method instead')
    async def advanced_sleep_mode(self) -> bool:
        """Set advanced sleep mode."""
        return await self.set_mode(FanModes.SLEEP)

    @deprecated('Use `set_sleep_mode` method instead')
    async def sleep_mode(self) -> bool:
        """Adapter to set advanced sleep mode."""
        return await self.set_advanced_sleep_mode()

    @deprecated('Use `set_mode` method instead')
    async def mode_toggle(self, mode: str) -> bool:
        """Set mode to specified mode."""
        return await self.set_mode(mode)



"""Base data models for API requests and response.

These models are used to define the structure of the requests and
responses from the API. They use Mashumaro for serialization and
deserialization. The `DataClassConfigMixin` class sets default options
for `orjson` and `Mashumaro`.

Note:
    Dataclasses should follow the naming convention of
    Request/Response + API Name. Internal models can have any descriptive
    name.


All models should inherit `ResponseBaseModel` or `RequestBaseModel`. Use
[pyvesync.const][pyvesync.const] to set default values and import here.

Attributes are inherited from the [const][pyvesync.const] module for
default values.
"""

from __future__ import annotations

from dataclasses import dataclass
from time import time
from typing import ClassVar

import orjson
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.const import (
    APP_ID,
    APP_VERSION,
    BYPASS_HEADER_UA,
    CLIENT_TYPE,
    DEFAULT_LANGUAGE,
    DEFAULT_REGION,
    DEFAULT_TZ,
    MOBILE_ID,
    PHONE_BRAND,
    PHONE_OS,
    TERMINAL_ID,
    USER_TYPE,
)

RequestHeaders = {
    'User-Agent': BYPASS_HEADER_UA,
    'Content-Type': 'application/json; charset=UTF-8',
}


class BaseModelConfig(BaseConfig):
    """Base config for dataclasses."""

    orjson_options = orjson.OPT_NON_STR_KEYS


@dataclass
class RequestBaseModel(DataClassORJSONMixin):
    """Base request model for API requests.

    Forbids extra keys in the request JSON.
    """

    class Config(BaseModelConfig):
        """orjson config for dataclasses."""

        forbid_extra_keys = True
        orjson_options = orjson.OPT_NON_STR_KEYS


@dataclass
class ResponseBaseModel(DataClassORJSONMixin):
    """Base response model for API responses.

    Allows extra keys in the response model and
    non-string keys in the JSON.
    """

    class Config(BaseConfig):
        """Config for dataclasses."""

        orjson_options = orjson.OPT_NON_STR_KEYS
        forbid_extra_keys = False


@dataclass
class DefaultValues:
    """Default request fields.

    Attributes for the default values of the request fields
    and static methods for preparing calculated fields.
    """

    _call_number: ClassVar[int] = 0
    userType: str = USER_TYPE
    appVersion: str = APP_VERSION
    clientType: str = CLIENT_TYPE
    appId: str = APP_ID
    phoneBrand: str = PHONE_BRAND
    phoneOS: str = PHONE_OS
    mobileId: str = MOBILE_ID
    deviceRegion: str = DEFAULT_REGION
    countryCode: str = DEFAULT_REGION
    userCountryCode: str = DEFAULT_REGION
    acceptLanguage: str = DEFAULT_LANGUAGE
    timeZone: str = DEFAULT_TZ
    terminalId: str = TERMINAL_ID
    debugMode: bool = False

    @staticmethod
    def traceId() -> str:
        """Trace ID CSRF token."""
        return str(int(time()))

    @staticmethod
    def newTraceId() -> str:
        """Generate a new trace ID."""
        DefaultValues._call_number += 1
        return f'APP{TERMINAL_ID[-5:-1]}{int(time())}-{DefaultValues._call_number:0>5}'


@dataclass
class ResponseCodeModel(ResponseBaseModel):
    """Model for the 'result' field in response."""

    traceId: str
    code: int
    msg: str | None



"""The Griffe extension."""

from __future__ import annotations

import contextlib
import logging
from typing import TYPE_CHECKING, Any

from griffe import (
    AliasResolutionError,
    Docstring,
    DocstringSectionText,
    Extension,
    GriffeLoader,
)

if TYPE_CHECKING:
    from griffe import Module, Object


logger = logging.getLogger(__name__)


def _docstring_above(obj: Object) -> str | None:
    with contextlib.suppress(IndexError, KeyError):
        for parent in obj.parent.mro():  # type: ignore[union-attr]
            if obj.name in parent.members and not parent.members[obj.name].is_alias:
                # Use parent of the parent object to avoid linking private methods
                return (
                    f'Inherited From [`{parent.members[obj.name].parent.name}`]'  # type: ignore[union-attr]
                    f'[{parent.members[obj.name].parent.canonical_path}]'  # type: ignore[union-attr]
                )
    return None


def _inherit_docstrings(  # noqa: C901
    obj: Object, loader: GriffeLoader, *, seen: set[str] | None = None
) -> None:
    if seen is None:
        seen = set()

    # if obj.path in seen:
    #     return  # noqa: ERA001

    seen.add(obj.path)

    if obj.is_module:
        for member in obj.members.values():
            if not member.is_alias:
                with contextlib.suppress(AliasResolutionError):
                    _inherit_docstrings(member, loader, seen=seen)  # type: ignore[arg-type]

    elif obj.is_class:
        for member in obj.all_members.values():
            if docstring_above := _docstring_above(member):  # type: ignore[arg-type]
                if member.docstring is None:
                    member.docstring = Docstring(
                        '',
                        parent=member,  # type: ignore[arg-type]
                        parser=loader.docstring_parser,
                        parser_options=loader.docstring_options,
                    )
                sections = member.docstring.parsed

                # Prevent inserting duplicate docstrings
                if (
                    sections
                    and sections[0].kind == 'text'
                    and sections[0].value.startswith('Inherited From')
                ):
                    continue
                sections.insert(0, DocstringSectionText(docstring_above))
                # This adds the Inherited notation to the base class, can't
                # figure out why.
            if member.is_class:
                _inherit_docstrings(member, loader, seen=seen)  # type: ignore[arg-type]


class InheritedNotation(Extension):
    """Griffe extension for inheriting docstrings."""

    def on_package(
        self,
        *,
        pkg: Module,
        loader: GriffeLoader,
        **kwargs: Any,  # noqa: ANN401, ARG002
    ) -> None:
        """Inherit docstrings from parent classes once the whole package is loaded."""
        _inherit_docstrings(pkg, loader)



"""This is a WIP, not implemented yet."""

from __future__ import annotations

import asyncio
import logging
from dataclasses import fields
from typing import TYPE_CHECKING

from pyvesync.models.home_models import (
    IntResponseHomeListModel,
    IntResponseHomeResultModel,
    RequestHomeModel,
    ResponseHomeModel,
)
from pyvesync.utils.errors import ErrorCodes, VeSyncAPIResponseError
from pyvesync.utils.helpers import Helpers

if TYPE_CHECKING:
    from pyvesync.base_devices import VeSyncBaseDevice
    from pyvesync.models.base_models import RequestBaseModel
    from pyvesync.vesync import VeSync


_LOGGER = logging.getLogger(__name__)


class VeSyncHome:
    """Class to handle home-related operations in VeSync."""

    def __init__(self, home_id: int, name: str, nickname: str | None = None) -> None:
        """Initialize the VeSyncHome instance."""
        self.home_id: int = home_id
        self.name: str = name
        self.nickname: str | None = nickname
        self.rooms: list[VeSyncRoom] = []

    @property
    def devices(self) -> list[VeSyncBaseDevice]:
        """Return a list of all devices in the home."""
        devices = []
        for room in self.rooms:
            devices.extend(room.devices)
        return devices

    async def update_devices(self, devices: list[VeSyncBaseDevice]) -> None:
        """Update the devices in the home."""
        update_tasks = [device.update() for device in devices]
        for update_coro in asyncio.as_completed(update_tasks):
            try:
                await update_coro
            except Exception as e:  # pylint: disable=broad-except
                _LOGGER.debug('Error updating device %s', e)

    @staticmethod
    def _build_request_model(
        manager: VeSync, request_model: type[RequestBaseModel]
    ) -> RequestBaseModel:
        """Build the request model for home data."""
        req_fields = tuple(field.name for field in fields(request_model) if field.init)
        body = Helpers.get_defaultvalues_attributes(req_fields)
        body.update(Helpers.get_manager_attributes(manager, req_fields))
        return request_model(**body)

    @classmethod
    def _process_home_list(
        cls, home_list: list[IntResponseHomeListModel]
    ) -> list[VeSyncHome]:
        """Process the home list and return a list of VeSyncHome instances."""
        homes = []
        for home in home_list:
            if not isinstance(home, IntResponseHomeListModel):
                msg = (
                    'Invalid home list item type.'
                    f'Expected IntResponseHomeListModel, got {home}'
                )
                raise VeSyncAPIResponseError(msg)
            homes.append(VeSyncHome(home.homeId, home.homeName, home.nickname))
        return homes

    @classmethod
    async def build_homes(cls, manager: VeSync) -> bool:
        """Get home information.

        This method retrieves the home list from the VeSync API and populates
        the VeSyncHome instances in the manager. The objects are stored in
        `manager.homes` attribute as a list.

        Args:
            manager (VeSync): The VeSync instance to use for the API call.

        Returns:
            bool: True if the home list was successfully retrieved and populated.

        Raises:
            VeSyncAPIResponseError: If the API response contains an error or
                if the home list is empty.
        """
        body = cls._build_request_model(manager, RequestHomeModel)
        response, _ = await manager.async_call_api(
            '/cloud/v1/homeManaged/getHomeList', method='post', json_object=body
        )
        if response is None:
            raise VeSyncAPIResponseError(
                'Response is None, enable debugging to see more information.'
            )

        resp_model = ResponseHomeModel.from_dict(response)
        if resp_model.code != 0:
            error = ErrorCodes.get_error_info(resp_model.code)
            if resp_model.msg is not None:
                error.message = f'{resp_model.msg} ({error.message})'

            msg = f'Failed to get home list with error: {error.to_json()}'
            raise VeSyncAPIResponseError(msg)
        result = resp_model.result
        if not isinstance(result, IntResponseHomeResultModel):
            msg = (
                'Error in home list API response.'
                f'Expected IntResponseHomeResultModel, got {result}'
            )
            raise VeSyncAPIResponseError(msg)
        home_list = result.homeList
        if not home_list:
            raise VeSyncAPIResponseError('No homes found in the response.')
        for home in home_list:
            if not isinstance(home, IntResponseHomeListModel):
                msg = (
                    'Invalid home list item type.'
                    f'Expected IntResponseHomeListModel, got {home}'
                )
                raise VeSyncAPIResponseError(msg)
        return True


class VeSyncRoom:
    """Class to handle room-related operations in VeSync."""

    def __init__(self, room_id: str, name: str) -> None:
        """Initialize the VeSyncRoom instance."""
        self.room_id: str = room_id
        self.name: str = name
        self.devices: list[VeSyncBaseDevice] = []

    @staticmethod
    def _build_request_model(
        manager: VeSync, request_model: type[RequestBaseModel]
    ) -> RequestBaseModel:
        """Build the request model for room data."""
        req_fields = tuple(field.name for field in fields(request_model) if field.init)
        body = Helpers.get_defaultvalues_attributes(req_fields)
        body.update(Helpers.get_manager_attributes(manager, req_fields))
        return request_model(**body)



"""
Switch Device API Responses

SWITCHES variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

"""
from copy import deepcopy
# from pyvesync.devices import vesyncswitch
from pyvesync.device_map import switch_modules
from pyvesync.const import DeviceStatus, ConnectionStatus
from defaults import TestDefaults, FunctionResponsesV1, build_bypass_v1_response

SWITCHES = [m.setup_entry for m in switch_modules]
SWITCHES_NUM = len(SWITCHES)


class SwitchDefaults:
    device_status = DeviceStatus.ON
    connection_status = ConnectionStatus.ONLINE
    indicator_status = DeviceStatus.ON
    rgb_status = DeviceStatus.ON


SWITCH_DETAILS: dict[str, dict] = {
    "ESWL01": {  # V1
        "deviceStatus": SwitchDefaults.device_status.value,
        "activeTime": TestDefaults.active_time,
        "deviceName": "Etekcity Light Switch",
        "deviceImg": "",
        "connectionStatus": SwitchDefaults.connection_status.value,
    },
    "ESWL03": {  # V1
        "deviceStatus": SwitchDefaults.device_status.value,
        "activeTime": TestDefaults.active_time,
        "deviceName": "Etekcity Light Switch",
        "deviceImg": "",
        "connectionStatus": SwitchDefaults.connection_status.value,
    },
    "ESWD16": {
        "deviceStatus": SwitchDefaults.device_status.value,
        "activeTime": TestDefaults.active_time,
        "devicename": "Etekcity Dimmer Switch",
        "indicatorlightStatus": SwitchDefaults.indicator_status.value,
        "startMode": None,
        "brightness": TestDefaults.brightness,
        "rgbStatus": SwitchDefaults.rgb_status.value,
        "rgbValue": TestDefaults.color.rgb.to_dict(),
        "timer": None,
        "schedule": None,
        "deviceImg": "https://image.vesync.com/defaultImages/deviceDefaultImages/wifiwalldimmer_240.png",
        "connectionStatus": SwitchDefaults.connection_status.value,
    },
}


# class SwitchDetails:
#     details_ws = (
#         {
#             'code': 0,
#             'msg': None,
#             'deviceStatus': 'on',
#             'connectionStatus': 'online',
#             'activeTime': Defaults.active_time,
#             'power': 'None',
#             'voltage': 'None',
#         },
#         200,
#     )
#     details_eswd16 = ({
#                           "code": 0,
#                           "msg": "请求成功",
#                           "traceId": Defaults.trace_id,
#                           "indicatorlightStatus": "on",
#                           "timer": None,
#                           "schedule": None,
#                           "brightness": "100",
#                           "startMode": None,
#                           "activeTime": Defaults.active_time,
#                           "rgbStatus": "on",
#                           "rgbValue": {
#                               "red": Defaults.color.rgb.red,
#                               "blue": Defaults.color.rgb.blue,
#                               "green": Defaults.color.rgb.green
#                           },
#                           "connectionStatus": "online",
#                           "devicename": Defaults.name('ESWD16'),
#                           "deviceStatus": "on"
#                       }, 200)


DETAILS_RESPONSES = {
    'ESWL01': build_bypass_v1_response(result_dict=SWITCH_DETAILS['ESWL01']),
    'ESWD16': build_bypass_v1_response(result_dict=SWITCH_DETAILS['ESWD16']),
    'ESWL03': build_bypass_v1_response(result_dict=SWITCH_DETAILS['ESWL03']),
}

METHOD_RESPONSES = {
    'ESWL01': deepcopy(FunctionResponsesV1),
    'ESWD16': deepcopy(FunctionResponsesV1),
    'ESWL03': deepcopy(FunctionResponsesV1),
}



"""Test VeSync login method."""
# from aiohttp.web_response import Response
from unittest.mock import MagicMock
import pytest
import orjson

from pyvesync.utils.errors import VeSyncLoginError
from pyvesync import const
from pyvesync.utils.helpers import Helpers
from pyvesync.models.vesync_models import (
    ResponseLoginModel,
    RequestLoginTokenModel,
    RequestGetTokenModel,
)


from aiohttp_mocker import AiohttpMockSession
from base_test_cases import TestApiFunc, TestBase
import call_json
from defaults import TestDefaults


TOKEN_ERROR_CODE = -11001000
TOKEN_ERROR_RESP = call_json.response_body(TOKEN_ERROR_CODE, "Token expired")

US_BASE_URL = const.API_BASE_URL_US
EU_BASE_URL = const.API_BASE_URL_EU

DEFAULT_ENDPOINT = '/endpoint'


INVALID_PASSWORD_ERROR = -11201000
INVALID_PASSWORD_RESP = {'code': INVALID_PASSWORD_ERROR, 'msg': 'success'}


LOGIN_RET_BODY = {
    "traceId": TestDefaults.trace_id,
    "code": 0,
    "msg": "",
    "stacktrace": None,
    "result": {
        "accountID": TestDefaults.account_id,
        "avatarIcon": "",
        "acceptLanguage": "en",
        "gdprStatus": True,
        "nickName": "nick",
        "userType": "1",
        "token": TestDefaults.token,
        "countryCode": TestDefaults.country_code,
    },
}

LOGIN_REQUESTS = call_json.LoginRequests
LOGIN_RESPONSES = call_json.LoginResponses
LOGIN_ENDPOINT = '/globalPlatform/api/accountAuth/v1/authByPWDOrOTM'
LOGIN_TOKEN_ENDPOINT = '/user/api/accountManage/v1/loginByAuthorizeCode4Vesync'

VARIABLE_FIELDS = {
    'traceId': TestDefaults.trace_id,
    'appId': TestDefaults.app_id,
    'terminalId': TestDefaults.terminal_id,
}

request_login_token_model = RequestLoginTokenModel(
    authorizeCode=TestDefaults.authorization_code,
    method='loginByAuthorizeCode4Vesync',
    userCountryCode=TestDefaults.country_code,
    traceId=TestDefaults.trace_id,
    terminalId=TestDefaults.terminal_id,
)
request_get_token_model = RequestGetTokenModel(
    email=TestDefaults.email,
    password=TestDefaults.password,
    method='authByPWDOrOTM',
    userCountryCode=TestDefaults.country_code,
    traceId=TestDefaults.trace_id,
    terminalId=TestDefaults.terminal_id,
    appID=TestDefaults.app_id,
)
response_get_token_model = ResponseLoginModel.from_dict(LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS)

response_login_token_model = ResponseLoginModel.from_dict(LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS)


class TestLoginModels(TestBase):
    """Test VeSync login response models."""

    def test_login_success(self):
        """Test login response model."""
        self.mock_api.side_effect = [
            (response_get_token_model.to_dict(), 200),
            (response_login_token_model.to_dict(), 200)
        ]
        self.run_in_loop(self.manager.login)
        assert self.mock_api.call_count == 2
        assert request_get_token_model.email == TestDefaults.email
        assert request_get_token_model.password == Helpers.hash_password(TestDefaults.password)
        assert self.manager.token == TestDefaults.token
        assert self.manager.account_id == TestDefaults.account_id
        call_list = self.mock_api.call_args_list
        assert call_list[0].args[0] == LOGIN_ENDPOINT
        assert call_list[0].args[1] == 'post'
        assert call_list[1].args[0] == LOGIN_TOKEN_ENDPOINT
        assert call_list[1].args[1] == 'post'
        get_token_kwargs = call_list[0].kwargs['json_object']
        get_token_kwargs.appID = TestDefaults.app_id
        get_token_kwargs.terminalId = TestDefaults.terminal_id
        get_token_kwargs.traceId = TestDefaults.trace_id
        assert get_token_kwargs == request_get_token_model
        login_token_kwargs = call_list[1].kwargs['json_object']
        login_token_kwargs.terminalId = TestDefaults.terminal_id
        login_token_kwargs.traceId = TestDefaults.trace_id
        assert login_token_kwargs == request_login_token_model


class TestLogin(TestApiFunc):
    """Test VeSync login class."""
    def test_invalid_password(self):
        """Test login with invalid user/password."""
        self.mock_api.return_value.request.return_value = AiohttpMockSession(
            method='post',
            url=US_BASE_URL + self.default_endpoint,
            status=200,
            response=orjson.dumps(INVALID_PASSWORD_RESP),
        )
        with pytest.raises(VeSyncLoginError):
            self.run_in_loop(self.manager.login)

    def _build_token_request(self):
        request_get_token_model = RequestGetTokenModel(
            email=TestDefaults.email,
            password=TestDefaults.password,
            method='authByPWDOrOTM',
            userCountryCode=TestDefaults.country_code,
            traceId=TestDefaults.trace_id,
            terminalId=TestDefaults.terminal_id,
            appID=TestDefaults.app_id,
        )
        return request_get_token_model

    def test_token_expired(self):
        """Test login with expired token."""
        token_error_response = orjson.dumps(TOKEN_ERROR_RESP)
        get_token_endpoint = LOGIN_ENDPOINT
        login_token_endpoint = LOGIN_TOKEN_ENDPOINT
        final_response = call_json.DeviceList.device_list_response()
        first_final_endpoint = '/cloud/v1/deviceManaged/devices'
        get_token_response = orjson.dumps(LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS)
        login_token_response = orjson.dumps(LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS)
        self.mock_api.return_value = MagicMock()
        self.mock_api.return_value.request.side_effect = [
            AiohttpMockSession(
                method='post',
                url=US_BASE_URL + first_final_endpoint,
                status=200,
                response=token_error_response,
            ),
            AiohttpMockSession(
                method='post',
                url=US_BASE_URL + get_token_endpoint,
                status=200,
                response=get_token_response,
            ),
            AiohttpMockSession(
                method='post',
                url=US_BASE_URL + login_token_endpoint,
                status=200,
                response=login_token_response,
            ),
            AiohttpMockSession(
                method='post',
                url=US_BASE_URL + first_final_endpoint,
                status=200,
                response=orjson.dumps(final_response),
            ),
        ]
        self.run_in_loop(self.manager.get_devices)
        assert len(self.mock_api.mock_calls) == 5  # Includes __init__ call
        call_list = self.mock_api.mock_calls
        assert call_list[1].args[0] == 'post'
        assert call_list[1].kwargs['url'] == US_BASE_URL + first_final_endpoint
        assert call_list[2].args[0] == 'post'
        assert call_list[2].kwargs['url'] == US_BASE_URL + get_token_endpoint
        assert call_list[3].args[0] == 'post'
        assert call_list[3].kwargs['url'] == US_BASE_URL + login_token_endpoint
        assert call_list[4].args[0] == 'post'
        assert call_list[4].kwargs['url'] == US_BASE_URL + first_final_endpoint



"""Data models for VeSync switches.

These models inherit from `ResponseBaseModel` and `RequestBaseModel` from the
`base_models` module.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import orjson
from mashumaro.config import BaseConfig

from pyvesync.models.base_models import ResponseBaseModel, ResponseCodeModel
from pyvesync.models.bypass_models import RequestBypassV1


@dataclass
class ResponseSwitchDetails(ResponseCodeModel):
    """Dimmer and Wall Switch Details Response Dict."""

    result: InternalDimmerDetailsResult | InternalSwitchResult | None = None


@dataclass
class InternalSwitchResult(ResponseBaseModel):
    """Dimmer Status Response Dict."""

    deviceStatus: str
    connectionStatus: str
    activeTime: int


@dataclass
class InternalDimmerDetailsResult(ResponseBaseModel):
    """Dimmer Details Result Dict."""

    devicename: str
    brightness: int
    indicatorlightStatus: str
    rgbStatus: str
    rgbValue: DimmerRGB
    deviceStatus: str
    connectionStatus: str
    activeTime: int = 0
    timer: Any | None = None
    schedule: Any | None = None
    deviceImg: str | None = None


@dataclass
class DimmerRGB:
    """Dimmer RGB Color Dict."""

    red: int
    green: int
    blue: int


@dataclass
class RequestSwitchBase(RequestBypassV1):
    """Base Dimmer Request Dict.

    Inherits from RequestBypassV1 to include the common fields for all requests.
    """


@dataclass
class RequestDimmerBrightness(RequestBypassV1):
    """Dimmer Status Request Dict."""

    brightness: str


@dataclass
class RequestDimmerDetails(RequestBypassV1):
    """Dimmer Details Request Dict."""


@dataclass
class RequestSwitchStatus(RequestBypassV1):
    """Dimmer Details Request Dict."""

    status: str
    switchNo: int


@dataclass
class RequestDimmerStatus(RequestBypassV1):
    """Dimmer Status Request Dict."""

    status: str
    rgbValue: dict | None = None

    class Config(BaseConfig):  # type: ignore[override]
        """Dimmer Indicator Control Config Dict."""

        omit_none = True
        omit_default = True
        orjson_options = orjson.OPT_NON_STR_KEYS



"""VeSync Humidifier Devices."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import orjson
from typing_extensions import deprecated

from pyvesync.base_devices.humidifier_base import BreathingLampState, VeSyncHumidifier
from pyvesync.const import ConnectionStatus, DeviceStatus, DryingModes
from pyvesync.models import humidifier_models as models
from pyvesync.models.bypass_models import (
    ResultV2GetTimer,
    ResultV2GetTimerV2,
    ResultV2SetTimer,
)
from pyvesync.utils.device_mixins import BypassV2Mixin, process_bypassv2_result
from pyvesync.utils.helpers import Helpers, Timer, Validators

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import HumidifierMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class VeSyncHumid200300S(BypassV2Mixin, VeSyncHumidifier):
    """300S Humidifier Class.

    Primary class for VeSync humidifier devices.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response info from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize 200S/300S Humidifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: models.ClassicLVHumidResult) -> None:
        """Set state from get_details API model."""
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = DeviceStatus.from_bool(resp_model.enabled)
        self.state.mode = self._reverse_mist_modes.get(resp_model.mode)
        self.state.humidity = resp_model.humidity
        self.state.mist_virtual_level = resp_model.mist_virtual_level or 0
        self.state.mist_level = resp_model.mist_level or 0
        self.state.water_lacks = resp_model.water_lacks
        self.state.humidity_high = resp_model.humidity_high
        self.state.water_tank_lifted = resp_model.water_tank_lifted
        self.state.auto_stop_target_reached = resp_model.automatic_stop_reach_target
        if self.supports_nightlight and resp_model.night_light_brightness is not None:
            self.state.nightlight_brightness = resp_model.night_light_brightness
            self.state.nightlight_status = (
                DeviceStatus.ON
                if resp_model.night_light_brightness > 0
                else DeviceStatus.OFF
            )
        self.state.display_status = DeviceStatus.from_bool(resp_model.display)
        if self.supports_warm_mist and resp_model.warm_level is not None:
            self.state.warm_mist_level = resp_model.warm_level
            self.state.warm_mist_enabled = resp_model.warm_enabled
        config = resp_model.configuration
        if config is not None:
            self.state.auto_target_humidity = config.auto_target_humidity
            self.state.automatic_stop_config = config.automatic_stop
            self.state.display_set_status = DeviceStatus.from_bool(config.display)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getHumidifierStatus')
        r_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, models.ClassicLVHumidResult
        )
        if r_model is None:
            return
        self._set_state(r_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.ON

        payload_data = {'enabled': toggle, 'id': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> Timer | None:
        r_dict = await self.call_bypassv2_api('getTimer')
        result_model = process_bypassv2_result(
            self, logger, 'get_timer', r_dict, ResultV2GetTimer
        )
        if result_model is None:
            return None
        if not result_model.timers:
            logger.debug('No timers found')
            return None
        timer = result_model.timers[0]
        self.state.timer = Timer(
            timer_duration=timer.total,
            action=timer.action,
            id=timer.id,
        )
        return self.state.timer

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer to clear, run get_timer() first.')
            return False
        payload = {
            'id': self.state.timer.id,
        }
        r_dict = await self.call_bypassv2_api('delTimer', payload)
        r = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r is None:
            return False
        self.state.timer = None
        return True

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.OFF
                if self.state.device_status == DeviceStatus.ON
                else DeviceStatus.ON
            )
        payload_data = {
            'action': str(action),
            'total': duration,
        }
        r_dict = await self.call_bypassv2_api('addTimer', payload_data)
        r = process_bypassv2_result(self, logger, 'set_timer', r_dict, ResultV2SetTimer)
        if r is None:
            return False

        self.state.timer = Timer(
            timer_duration=duration, action=action, id=r.id, remaining=0
        )
        return True

    @deprecated('Use turn_on_automatic_stop() instead.')
    async def automatic_stop_on(self) -> bool:
        """Turn 200S/300S Humidifier automatic stop on."""
        return await self.toggle_automatic_stop(True)

    @deprecated('Use turn_off_automatic_stop() instead.')
    async def automatic_stop_off(self) -> bool:
        """Turn 200S/300S Humidifier automatic stop on."""
        return await self.toggle_automatic_stop(False)

    @deprecated('Use toggle_automatic_stop(toggle: bool) instead.')
    async def set_automatic_stop(self, mode: bool) -> bool:
        """Set 200S/300S Humidifier to automatic stop."""
        return await self.toggle_automatic_stop(mode)

    async def toggle_automatic_stop(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.automatic_stop_config is not True

        payload_data = {'enabled': toggle}
        r_dict = await self.call_bypassv2_api('setAutomaticStop', payload_data)
        r = Helpers.process_dev_response(logger, 'set_automatic_stop', self, r_dict)
        if r is None:
            return False

        self.state.automatic_stop_config = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    @deprecated('Use toggle_display(toggle: bool) instead.')
    async def set_display(self, toggle: bool) -> bool:
        """Deprecated method to toggle display on/off.

        Use toggle_display(toggle: bool) instead.
        """
        return await self.toggle_display(toggle)

    async def toggle_display(self, toggle: bool) -> bool:
        payload_data = {'state': toggle}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)
        r = Helpers.process_dev_response(logger, 'set_display', self, r_dict)
        if r is None:
            return False
        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.display_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_humidity(self, humidity: int) -> bool:
        if not Validators.validate_range(humidity, *self.target_minmax):
            logger.warning(
                'Invalid humidity, must be between %s and %s', *self.target_minmax
            )
            return False

        payload_data = {'target_humidity': humidity}
        r_dict = await self.call_bypassv2_api('setTargetHumidity', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity', self, r_dict)
        if r is None:
            return False
        self.state.auto_target_humidity = humidity
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_nightlight_brightness(self, brightness: int) -> bool:
        if not self.supports_nightlight:
            logger.warning(
                '%s is a %s does not have a nightlight or it is not supported.',
                self.device_name,
                self.device_type,
            )
            return False

        if not Validators.validate_zero_to_hundred(brightness):
            logger.warning('Brightness value must be set between 0 and 100')
            return False

        payload_data = {'night_light_brightness': brightness}
        r_dict = await self.call_bypassv2_api('setNightLightBrightness', payload_data)
        r = Helpers.process_dev_response(
            logger, 'set_night_light_brightness', self, r_dict
        )
        if r is None:
            return False
        self.state.nightlight_brightness = brightness
        self.state.nightlight_status = (
            DeviceStatus.ON if brightness > 0 else DeviceStatus.OFF
        )
        return True

    async def toggle_nightlight(self, toggle: bool | None = None) -> bool:
        if not self.supports_nightlight:
            logger.warning(
                '%s is a %s does not have a nightlight or it is not supported.',
                self.device_name,
                self.device_type,
            )
            return False

        if toggle is None:
            toggle = self.state.nightlight_status != DeviceStatus.ON
        brightness = 100 if toggle else 0
        return await self.set_nightlight_brightness(brightness)

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.mist_modes:
            logger.warning('Invalid humidity mode used - %s', mode)
            logger.info(
                'Proper modes for this device are - %s',
                orjson.dumps(
                    self.mist_modes, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
                ),
            )
            return False

        payload_data = {'mode': self.mist_modes[mode]}
        r_dict = await self.call_bypassv2_api('setHumidityMode', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_warm_level(self, warm_level: int) -> bool:
        if not self.supports_warm_mist:
            logger.debug(
                '%s is a %s does not have a mist warmer',
                self.device_name,
                self.device_type,
            )
            return False

        if warm_level not in self.warm_mist_levels:
            logger.warning('warm_level value must be - %s', str(self.warm_mist_levels))
            return False

        payload_data = {'type': 'warm', 'level': warm_level, 'id': 0}
        r_dict = await self.call_bypassv2_api('setVirtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_warm_level', self, r_dict)
        if r is None:
            return False

        self.state.warm_mist_level = warm_level
        self.state.warm_mist_enabled = True
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mist_level(self, level: int) -> bool:
        if level not in self.mist_levels:
            logger.warning(
                'Humidifier mist level must be between %s and %s',
                self.mist_levels[0],
                self.mist_levels[-1],
            )
            return False

        payload_data = {'id': 0, 'level': level, 'type': 'mist'}
        r_dict = await self.call_bypassv2_api('setVirtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mist_level', self, r_dict)
        if r is None:
            return False

        self.state.mist_virtual_level = level
        self.state.mist_level = level
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncHumid200S(VeSyncHumid200300S):
    """Levoit Classic 200S Specific class.

    Overrides the `toggle_display(toggle: bool)` method of the
    VeSyncHumid200300S class.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize levoit 200S device class.

        This overrides the `toggle_display(toggle: bool)` method of the
        VeSyncHumid200300S class.
        """
        super().__init__(details, manager, feature_map)

    async def toggle_display(self, toggle: bool) -> bool:
        payload_data = {'enabled': toggle, 'id': 0}
        r_dict = await self.call_bypassv2_api('setIndicatorLightSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.display_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncSuperior6000S(BypassV2Mixin, VeSyncHumidifier):
    """Superior 6000S Humidifier.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize Superior 6000S Humidifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: models.Superior6000SResult) -> None:
        """Set state from Superior 6000S API result model."""
        self.state.device_status = DeviceStatus.from_int(resp_model.powerSwitch)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.mode = self._reverse_mist_modes.get(resp_model.workMode)
        if self.state.mode is None:
            logger.warning('Unknown mist mode received: %s', resp_model.workMode)

        self.state.auto_target_humidity = resp_model.targetHumidity
        self.state.humidity = resp_model.humidity
        self.state.mist_level = resp_model.mistLevel
        self.state.mist_virtual_level = resp_model.virtualLevel
        self.state.water_lacks = bool(resp_model.waterLacksState)
        self.state.water_tank_lifted = bool(resp_model.waterTankLifted)
        self.state.automatic_stop_config = bool(resp_model.autoStopSwitch)
        self.state.auto_stop_target_reached = bool(resp_model.autoStopState)
        self.state.display_set_status = DeviceStatus.from_int(resp_model.screenState)
        self.state.display_status = DeviceStatus.from_int(resp_model.screenState)
        self.state.auto_preference = resp_model.autoPreference
        self.state.filter_life = resp_model.filterLifePercent
        self.state.child_lock = bool(resp_model.childLockSwitch)
        self.state.temperature = (
            resp_model.temperature / 10
        )  # Fahrenheit but without decimals

        drying_mode = resp_model.dryingMode
        if drying_mode is not None:
            self.state.drying_mode_status = DryingModes.from_int(drying_mode.dryingState)
            self.state.drying_mode_auto_switch = DeviceStatus.from_int(
                drying_mode.autoDryingSwitch
            )
            self.state.drying_mode_level = drying_mode.dryingLevel
            self.state.drying_mode_time_remain = drying_mode.dryingRemain
        if resp_model.timerRemain > 0:
            self.state.timer = Timer(
                resp_model.timerRemain,
                DeviceStatus.from_bool(self.state.device_status != DeviceStatus.ON),
            )

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getHumidifierStatus')
        r_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, models.Superior6000SResult
        )
        if r_model is None:
            return

        self._set_state(r_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        payload_data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_automatic_stop(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.automatic_stop_config is not True

        payload_data = {'autoStopSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setAutoStopSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_automatic_stop', self, r_dict)
        if r is None:
            return False

        self.state.automatic_stop_config = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_drying_mode(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.drying_mode_status != DeviceStatus.ON

        payload_data = {'autoDryingSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDryingMode', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_drying_mode', self, r_dict)
        if r is None:
            return False

        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.drying_mode_auto_switch = DeviceStatus.from_bool(toggle)
        return True

    async def toggle_display(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.display_set_status != DeviceStatus.ON

        payload_data = {'screenSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)
        r = Helpers.process_dev_response(logger, 'set_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_humidity(self, humidity: int) -> bool:
        if not Validators.validate_range(humidity, *self.target_minmax):
            logger.warning('Humidity value must be set between 30 and 80')
            return False

        payload_data = {'targetHumidity': humidity}
        r_dict = await self.call_bypassv2_api('setTargetHumidity', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity', self, r_dict)
        if r is None:
            return False

        self.state.auto_target_humidity = humidity
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.mist_modes:
            logger.warning('Invalid humidity mode used - %s', mode)
            logger.info(
                'Proper modes for this device are - %s',
                orjson.dumps(
                    self.mist_modes, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
                ),
            )
            return False

        payload_data = {'workMode': self.mist_modes[mode]}
        r_dict = await self.call_bypassv2_api('setHumidityMode', payload_data)

        r = Helpers.process_dev_response(logger, 'set_humidity_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mist_level(self, level: int) -> bool:
        if level not in self.mist_levels:
            logger.warning('Humidifier mist level must be between 0 and 9')
            return False

        payload_data = {'levelIdx': 0, 'virtualLevel': level, 'levelType': 'mist'}
        r_dict = await self.call_bypassv2_api('setVirtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mist_level', self, r_dict)
        if r is None:
            return False

        self.state.mist_level = level
        self.state.mist_virtual_level = level
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.child_lock is True

        payload_data = {'childLockSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setChildLock', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_child_lock', self, r_dict)
        if r is None:
            return False

        self.state.child_lock = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncHumid1000S(VeSyncHumid200300S):
    """Levoit OasisMist 1000S Specific class.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize levoit 1000S device class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: models.InnerHumidifierBaseResult) -> None:
        """Set state of Levoit 1000S from API result model."""
        if not isinstance(resp_model, models.Levoit1000SResult):
            return
        self.state.device_status = DeviceStatus.from_int(resp_model.powerSwitch)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.mode = self._reverse_mist_modes.get(resp_model.workMode)
        self.state.humidity = resp_model.humidity
        self.state.auto_target_humidity = resp_model.targetHumidity
        self.state.mist_level = resp_model.mistLevel
        self.state.mist_virtual_level = resp_model.virtualLevel
        self.state.water_lacks = bool(resp_model.waterLacksState)
        self.state.water_tank_lifted = bool(resp_model.waterTankLifted)
        self.state.automatic_stop_config = bool(resp_model.autoStopSwitch)
        self.state.auto_stop_target_reached = bool(resp_model.autoStopState)
        self.state.display_set_status = DeviceStatus.from_int(resp_model.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(resp_model.screenState)
        if resp_model.nightLight is not None:
            self.state.nightlight_brightness = resp_model.nightLight.brightness
            self.state.nightlight_status = DeviceStatus.from_int(
                resp_model.nightLight.nightLightSwitch
            )

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getHumidifierStatus')
        r_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, models.Levoit1000SResult
        )
        if r_model is None:
            return

        self._set_state(r_model)

    @deprecated('Use toggle_display() instead.')
    async def set_display(self, toggle: bool) -> bool:
        """Toggle display on/off.

        This is a deprecated method, please use toggle_display() instead.
        """
        return await self.toggle_switch(toggle)

    async def toggle_display(self, toggle: bool) -> bool:
        payload_data = {'screenSwitch': int(toggle)}
        body = self._build_request('setDisplay', payload_data)
        headers = Helpers.req_header_bypass()

        r_dict, _ = await self.manager.async_call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            method='post',
            headers=headers,
            json_object=body.to_dict(),
        )
        r = Helpers.process_dev_response(logger, 'set_display', self, r_dict)
        if r is None:
            return False
        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.mist_modes:
            logger.warning('Invalid humidity mode used - %s', mode)
            logger.info(
                'Proper modes for this device are - %s',
                orjson.dumps(
                    self.mist_modes, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
                ),
            )
            return False

        payload_data = {'workMode': self.mist_modes[mode]}
        r_dict = await self.call_bypassv2_api('setHumidityMode', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode

        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mist_level(self, level: int) -> bool:
        if level not in self.mist_levels:
            logger.warning('Humidifier mist level out of range')
            return False

        payload_data = {'levelIdx': 0, 'virtualLevel': level, 'levelType': 'mist'}
        r_dict = await self.call_bypassv2_api('virtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mist_level', self, r_dict)
        if r is None:
            return False

        self.state.mist_level = level
        self.state.mist_virtual_level = level
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        payload_data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_humidity(self, humidity: int) -> bool:
        if not Validators.validate_range(humidity, *self.target_minmax):
            logger.warning(
                'Humidity value must be set between %s and %s',
                self.target_minmax[0],
                self.target_minmax[1],
            )
            return False

        payload_data = {'targetHumidity': humidity}
        r_dict = await self.call_bypassv2_api('setTargetHumidity', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity', self, r_dict)
        if r is None:
            return False

        self.state.auto_target_humidity = humidity
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_automatic_stop(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.automatic_stop_config is not True

        payload_data = {'autoStopSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setAutoStopSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'set_automatic_stop', self, r_dict)

        if r is None:
            return False
        self.state.automatic_stop_config = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_nightlight(self, toggle: bool | None = None) -> bool:
        if not self.supports_nightlight:
            logger.warning(
                '%s is a %s does not have a nightlight or it is not supported.',
                self.device_name,
                self.device_type,
            )
            return False

        if toggle is None:
            toggle = self.state.nightlight_status != DeviceStatus.ON

        payload_data = {'nightLightSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setNightLightStatus', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_nightlight', self, r_dict)
        if r is None:
            return False

        self.state.nightlight_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_nightlight_brightness(self, brightness: int) -> bool:
        if not self.supports_nightlight:
            logger.warning(
                '%s is a %s does not have a nightlight or it is not supported.',
                self.device_name,
                self.device_type,
            )
            return False

        if not Validators.validate_zero_to_hundred(brightness):
            logger.warning('Brightness value must be set between 0 and 100')
            return False

        payload_data = {
            'brightness': brightness,
            'nightLightSwitch': 1 if brightness > 0 else 0,
        }
        r_dict = await self.call_bypassv2_api('setLightStatus', payload_data)
        r = Helpers.process_dev_response(
            logger, 'set_night_light_brightness', self, r_dict
        )
        if r is None:
            return False

        self.state.nightlight_brightness = brightness
        self.state.nightlight_status = (
            DeviceStatus.ON if brightness > 0 else DeviceStatus.OFF
        )
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncSproutHumid(BypassV2Mixin, VeSyncHumidifier):
    """VeSync Sprout Humidifier class.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize Sprout Humidifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: models.SproutHumidifierResult) -> None:
        """Set state from Sprout Humidifier API result model."""
        self.state.device_status = DeviceStatus.from_int(resp_model.powerSwitch)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.humidity = resp_model.humidity
        self.state.auto_target_humidity = resp_model.targetHumidity
        self.state.mist_virtual_level = resp_model.virtualLevel
        self.state.mist_level = resp_model.mistLevel
        self.state.mode = self._reverse_mist_modes.get(resp_model.workMode)
        self.state.water_lacks = bool(resp_model.waterLacksState)
        self.state.water_tank_lifted = bool(resp_model.waterTankLifted)
        self.state.automatic_stop_config = bool(resp_model.autoStopSwitch)
        self.state.auto_stop_target_reached = bool(resp_model.autoStopState)
        self.state.display_set_status = DeviceStatus.from_int(resp_model.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(resp_model.screenState)
        self.state.auto_preference = resp_model.autoPreference
        self.state.water_lacks_drying_switch = DeviceStatus.from_int(
            resp_model.waterShortageDryingSwitch
        )
        self.state.child_lock = bool(resp_model.childLockSwitch)
        self.state.filter_life = resp_model.filterLifePercent
        self.state.hepa_filter_life = resp_model.hepaFilterLifePercent
        self.state.temperature = (
            resp_model.temperature / 10
        )  # Response needs to be divided by 10
        if resp_model.breathingLamp is not None:
            self.state.breathing_lamp = BreathingLampState.from_model(
                resp_model.breathingLamp
            )
        if resp_model.nightLight is not None:
            self.state.nightlight_brightness = resp_model.nightLight.brightness
            self.state.nightlight_status = DeviceStatus.from_int(
                resp_model.nightLight.nightLightSwitch
            )
            self.state.nightlight_color_temp = resp_model.nightLight.colorTemperature
        if resp_model.dryingMode is not None:
            drying_mode = resp_model.dryingMode
            self.state.drying_mode_status = DryingModes.from_int(drying_mode.dryingState)
            self.state.drying_mode_auto_switch = DeviceStatus.from_int(
                drying_mode.autoDryingSwitch
            )
            self.state.drying_mode_time_remain = drying_mode.dryingRemain
            self.state.drying_mode_level = drying_mode.dryingLevel

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getHumidifierStatus')
        r_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, models.SproutHumidifierResult
        )
        if r_model is None:
            return
        self._set_state(r_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        payload_data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.mist_modes:
            logger.warning('Invalid humidity mode used - %s', mode)
            logger.info(
                'Proper modes for this device are - %s',
                orjson.dumps(
                    self.mist_modes, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
                ),
            )
            return False
        payload_data = {'workMode': self.mist_modes[mode]}
        r_dict = await self.call_bypassv2_api('setHumidityMode', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mode', self, r_dict)
        if r is None:
            return False
        self.state.mode = mode
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.display_set_status != DeviceStatus.ON

        payload_data = {'screenSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)
        r = Helpers.process_dev_response(logger, 'set_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.display_status = DeviceStatus.from_bool(toggle)
        return True

    async def set_mist_level(self, level: int) -> bool:
        if level not in self.mist_levels:
            logger.warning('Humidifier mist level out of range')
            return False

        payload_data = {'levelIdx': 0, 'virtualLevel': level, 'levelType': 'mist'}
        r_dict = await self.call_bypassv2_api('setVirtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mist_level', self, r_dict)
        if r is None:
            return False

        self.state.mist_level = level
        self.state.mist_virtual_level = level
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_humidity(self, humidity: int) -> bool:
        if not Validators.validate_range(humidity, *self.target_minmax):
            logger.warning(
                'Humidity value must be set between %s and %s',
                self.target_minmax[0],
                self.target_minmax[1],
            )
            return False

        payload_data = {'targetHumidity': humidity}
        r_dict = await self.call_bypassv2_api('setTargetHumidity', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity', self, r_dict)
        if r is None:
            return False

        self.state.auto_target_humidity = humidity
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_drying_mode(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.drying_mode_status != DeviceStatus.ON

        payload_data = {'autoDryingSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDryingMode', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_drying_mode', self, r_dict)
        if r is None:
            return False

        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.drying_mode_auto_switch = DeviceStatus.from_bool(toggle)
        return True

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.child_lock is True

        payload_data = {'childLockSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setChildLock', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_child_lock', self, r_dict)
        if r is None:
            return False

        self.state.child_lock = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def _set_nightlight_state(
        self,
        toggle: bool,
        brightness: int | None = None,
        color_temp: int | None = None,
    ) -> bool:
        if brightness is not None and not Validators.validate_zero_to_hundred(brightness):
            logger.warning('Brightness value must be set between 0 and 100')
            return False

        if color_temp is not None and not Validators.validate_range(
            color_temp, 2000, 3500
        ):
            logger.warning('Color temperature must be set between 2000K and 3500K')
            return False

        if self.state.nightlight_color_temp is None:
            self.state.nightlight_color_temp = 3500  # Default color temp if not set

        payload_data = {
            'brightness': brightness or self.state.nightlight_brightness,
            'colorTemperature': color_temp or self.state.nightlight_color_temp,
            'nightLightSwitch': int(toggle),
        }
        r_dict = await self.call_bypassv2_api('setLightStatus', payload_data)
        r = Helpers.process_dev_response(logger, 'set_night_light_state', self, r_dict)
        if r is None:
            return False

        self.state.nightlight_brightness = brightness
        self.state.nightlight_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_automatic_stop(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.automatic_stop_config is not True

        payload_data = {'autoStopSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setAutoStopSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_automatic_stop', self, r_dict)
        if r is None:
            return False

        self.state.automatic_stop_config = toggle
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncLV600S(BypassV2Mixin, VeSyncHumidifier):
    """LV600S (LUH-A603S) Humidifier with warm mist support.

    This class handles the newer LV600S models that use the new API format
    with powerSwitch, workMode, etc. and include warm mist functionality.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The manager object for API calls.
        feature_map (HumidifierMap): The feature map for the device.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.
    """

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize LV600S Humidifier class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: models.LV600SResult) -> None:
        """Set state from LV600S API result model."""
        self.state.device_status = DeviceStatus.from_int(resp_model.powerSwitch)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.mode = self._reverse_mist_modes.get(resp_model.workMode)
        if self.state.mode is None:
            logger.warning('Unknown mist mode received: %s', resp_model.workMode)

        self.state.auto_target_humidity = resp_model.targetHumidity
        self.state.humidity = resp_model.humidity
        self.state.mist_level = resp_model.mistLevel
        self.state.mist_virtual_level = resp_model.virtualLevel
        self.state.water_lacks = bool(resp_model.waterLacksState)
        self.state.water_tank_lifted = bool(resp_model.waterTankLifted)
        self.state.automatic_stop_config = bool(resp_model.autoStopSwitch)
        self.state.auto_stop_target_reached = bool(resp_model.autoStopState)
        self.state.display_set_status = DeviceStatus.from_int(resp_model.screenSwitch)
        self.state.display_status = DeviceStatus.from_int(resp_model.screenState)
        # Warm mist support
        self.state.warm_mist_enabled = resp_model.warmPower
        self.state.warm_mist_level = resp_model.warmLevel

        if resp_model.timerRemain > 0:
            self.state.timer = Timer(
                resp_model.timerRemain,
                DeviceStatus.from_bool(self.state.device_status != DeviceStatus.ON),
            )

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getHumidifierStatus')
        r_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, models.LV600SResult
        )
        if r_model is None:
            return

        self._set_state(r_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON

        payload_data = {'powerSwitch': int(toggle), 'switchIdx': 0}
        r_dict = await self.call_bypassv2_api('setSwitch', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_humidity(self, humidity: int) -> bool:
        if not Validators.validate_range(humidity, *self.target_minmax):
            logger.warning(
                'Humidity value must be set between %s and %s',
                self.target_minmax[0],
                self.target_minmax[1],
            )
            return False

        payload_data = {'targetHumidity': humidity}
        r_dict = await self.call_bypassv2_api('setTargetHumidity', payload_data)
        r = Helpers.process_dev_response(logger, 'set_humidity', self, r_dict)
        if r is None:
            return False

        self.state.auto_target_humidity = humidity
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mode(self, mode: str) -> bool:
        if mode not in self.mist_modes:
            logger.warning('Invalid humidity mode used - %s', mode)
            logger.info(
                'Proper modes for this device are - %s',
                orjson.dumps(
                    self.mist_modes, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
                ),
            )
            return False

        payload_data = {'workMode': self.mist_modes[mode]}
        r_dict = await self.call_bypassv2_api('setHumidityMode', payload_data)

        r = Helpers.process_dev_response(logger, 'set_mode', self, r_dict)
        if r is None:
            return False

        self.state.mode = mode
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_mist_level(self, level: int) -> bool:
        if level not in self.mist_levels:
            logger.warning('Humidifier mist level must be between 1 and 9')
            return False

        payload_data = {'levelIdx': 0, 'virtualLevel': level, 'levelType': 'mist'}
        r_dict = await self.call_bypassv2_api('setVirtualLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_mist_level', self, r_dict)
        if r is None:
            return False

        self.state.mist_level = level
        self.state.mist_virtual_level = level
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_warm_level(self, warm_level: int) -> bool:
        """Set warm mist level (0 = off, 1-3 = warm levels)."""
        if not self.supports_warm_mist:
            logger.debug(
                '%s is a %s does not have a mist warmer',
                self.device_name,
                self.device_type,
            )
            return False

        if warm_level not in self.warm_mist_levels:
            logger.warning('Warm mist level must be one of %s', self.warm_mist_levels)
            return False

        payload_data = {
            'levelIdx': 0,
            'levelType': 'warm',
            'mistLevel': 0,
            'warmLevel': warm_level,
        }
        r_dict = await self.call_bypassv2_api('setLevel', payload_data)
        r = Helpers.process_dev_response(logger, 'set_warm_level', self, r_dict)
        if r is None:
            return False

        self.state.warm_mist_level = warm_level
        self.state.warm_mist_enabled = warm_level > 0
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def toggle_display(self, toggle: bool | None = None) -> bool:
        """Toggle display on/off."""
        if toggle is None:
            toggle = self.state.display_set_status != DeviceStatus.ON

        payload_data = {'screenSwitch': int(toggle)}
        r_dict = await self.call_bypassv2_api('setDisplay', payload_data)
        r = Helpers.process_dev_response(logger, 'toggle_display', self, r_dict)
        if r is None:
            return False

        self.state.display_set_status = DeviceStatus.from_bool(toggle)
        self.state.display_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> Timer | None:
        """Get timer for humidifier using V2 API."""
        r_dict = await self.call_bypassv2_api('getTimerV2', {})
        result_model = process_bypassv2_result(
            self, logger, 'get_timer', r_dict, ResultV2GetTimerV2
        )
        if result_model is None:
            return None
        if not result_model.timers:
            logger.debug('No timers found')
            return None
        timer = result_model.timers[0]
        action = DeviceStatus.OFF
        if timer.startAct:
            action = DeviceStatus.ON if timer.startAct[0].act == 1 else DeviceStatus.OFF
        self.state.timer = Timer(
            timer_duration=timer.total,
            action=action,
            id=timer.id,
            remaining=timer.remain,
        )
        return self.state.timer

    async def clear_timer(self) -> bool:
        """Clear timer for humidifier using V2 API."""
        timer_id = self.state.timer.id if self.state.timer else 1
        payload = {'id': timer_id}
        r_dict = await self.call_bypassv2_api('delTimerV2', payload)
        r = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r is None:
            return False
        self.state.timer = None
        return True

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        """Set timer for humidifier using V2 API.

        Args:
            duration: Timer duration in seconds.
            action: Action when timer expires. Defaults to toggle current state.
        """
        if action is None:
            action = (
                DeviceStatus.OFF
                if self.state.device_status == DeviceStatus.ON
                else DeviceStatus.ON
            )
        act_val = 1 if action == DeviceStatus.ON else 0

        payload_data = {
            'startAct': [{'type': 'powerSwitch', 'num': 0, 'act': act_val}],
            'total': duration,
        }
        r_dict = await self.call_bypassv2_api('addTimerV2', payload_data)
        result_model = process_bypassv2_result(
            self, logger, 'set_timer', r_dict, ResultV2SetTimer
        )
        if result_model is None:
            return False

        self.state.timer = Timer(
            timer_duration=duration,
            action=action,
            id=result_model.id,
            remaining=duration,
        )
        return True



"""Data structures for handling colors."""

from __future__ import annotations

import colorsys
import logging
from dataclasses import InitVar, dataclass

from pyvesync.utils.helpers import Validators

_LOGGER = logging.getLogger(__name__)


@dataclass
class RGB:
    """RGB color space dataclass, for internal use in `utils.colors.Color` dataclass.

    Does not perform any validation, it should not be used directly. Used as an
    attribute in the :obj:`pyvesync.helpers.Color` dataclass. This should only be
    used through the :obj:`Color` dataclass with the Color.from_rgb(red, green, blue)
    classmethod.

    Attributes:
        red (float): The red component of the RGB color.
        green (float): The green component of the RGB color.
        blue (float): The blue component of the RGB color.
    """

    red: float
    green: float
    blue: float

    def __post_init__(self) -> None:
        """Convert to int."""
        self.red = int(self.red)
        self.green = int(self.green)
        self.blue = int(self.blue)

    def __str__(self) -> str:
        """Return string representation."""
        return f'RGB({self.red}, {self.green}, {self.blue})'

    def __repr__(self) -> str:
        """Return string representation."""
        return f'RGB(red={self.red}, green={self.green}, blue={self.blue})'

    def to_tuple(self) -> tuple[float, float, float]:
        """Return RGB values as tuple."""
        return self.red, self.green, self.blue

    def to_dict(self) -> dict[str, float]:
        """Return RGB values as dict."""
        return {
            'red': self.red,
            'green': self.green,
            'blue': self.blue,
        }


@dataclass
class HSV:
    """HSV color space dataclass, for internal use in `utils.colors.Color` dataclass.

    Does not perform any validation and should not be used directly, only
    by the `Color` dataclass through the Color.from_hsv(hue, saturation, value)
    classmethod or Color.rgb_to_hsv(red, green, blue) method.

    Attributes:
        hue (float): The hue component of the color, typically in the range [0, 360).
        saturation (float): The saturation component of the color,
            typically in the range [0, 1].
        value (float): The value (brightness) component of the color,
            typically in the range [0, 1].
    """

    hue: float
    saturation: float
    value: float

    def __post_init__(self) -> None:
        """Convert to int."""
        self.hue = int(self.hue)
        self.saturation = int(self.saturation)
        self.value = int(self.value)

    def __str__(self) -> str:
        """Return string representation."""
        return f'HSV({self.hue}, {self.saturation}, {self.value})'

    def __repr__(self) -> str:
        """Return string representation."""
        return f'HSV(hue={self.hue}, saturation={self.saturation}, value={self.value})'

    def to_tuple(self) -> tuple[float, float, float]:
        """Return HSV values as tuple."""
        return self.hue, self.saturation, self.value

    def to_dict(self) -> dict[str, float]:
        """Return HSV values as dict."""
        return {
            'hue': self.hue,
            'saturation': self.saturation,
            'value': self.value,
        }


@dataclass
class Color:
    """Dataclass for color values.

    This class should be instantiated through the `from_rgb` or `from_hsv`
    classmethods. It will return a `Color` object with the appropriate color
    values in RGB and HSV.

    Args:
        color_object (HSV | RGB): Named tuple with color values.

    Attributes:
        hsv (namedtuple): hue (0-360), saturation (0-100), value (0-100)
            see [`HSV dataclass`][pyvesync.utils.colors.HSV]
        rgb (namedtuple): red (0-255), green (0-255), blue (0-255)
            see [`RGB dataclass`][pyvesync.utils.colors.RGB]
    """

    color_object: InitVar[HSV | RGB]

    def __post_init__(
        self,
        color_object: HSV | RGB,
    ) -> None:
        """Check HSV or RGB Values and create named tuples."""
        if isinstance(color_object, HSV):
            self.hsv = color_object
            self.rgb = self.hsv_to_rgb(*self.hsv.to_tuple())
        elif isinstance(color_object, RGB):
            self.rgb = color_object
            self.hsv = self.rgb_to_hsv(*self.rgb.to_tuple())

    def __str__(self) -> str:
        """Return string representation."""
        return f'Color(hsv={self.hsv}, rgb={self.rgb})'

    def __repr__(self) -> str:
        """Return string representation."""
        return f'Color(hsv={self.hsv}, rgb={self.rgb})'

    def as_dict(self) -> dict[str, dict]:
        """Return color values as dict."""
        return {
            'hsv': {
                'hue': self.hsv.hue,
                'saturation': self.hsv.saturation,
                'value': self.hsv.value,
            },
            'rgb': {
                'red': self.rgb.red,
                'green': self.rgb.green,
                'blue': self.rgb.blue,
            },
        }

    @classmethod
    def from_rgb(
        cls, red: float | None, green: float | None, blue: float | None
    ) -> Color | None:
        """Create Color instance from RGB values.

        Args:
            red (NUMERIC_STRICT): The red component of the color,
                typically in the range [0, 255].
            green (NUMERIC_STRICT): The green component of the color,
                typically in the range [0, 255].
            blue (NUMERIC_STRICT): The blue component of the color,
                typically in the range [0, 255].

        Returns:
            Color | None: A Color object with the appropriate color values in RGB and HSV,
                or None if the input values are invalid.
        """
        if not Validators.validate_rgb(red, green, blue):
            _LOGGER.debug('Invalid RGB values')
            return None
        return cls(RGB(float(red), float(green), float(blue)))  # type: ignore[arg-type]

    @classmethod
    def from_hsv(
        cls, hue: float | None, saturation: float | None, value: float | None
    ) -> Color | None:
        """Create Color instance from HSV values.

        Args:
            hue (float): The hue component of the color,
                in the range [0, 360).
            saturation (float): The saturation component of the color,
                typically in the range [0, 1].
            value (float): The value (brightness) component of the color,
                typically in the range [0, 1].

        Returns:
            Color | None: A Color object with the appropriate color values in RGB and HSV,
                or None if the input values are invalid.
        """
        if not Validators.validate_hsv(hue, saturation, value):
            _LOGGER.debug('Invalid HSV values')
            return None
        return cls(
            HSV(float(hue), float(saturation), float(value))  # type: ignore[arg-type]
        )

    @staticmethod
    def hsv_to_rgb(hue: float, saturation: float, value: float) -> RGB:
        """Convert HSV to RGB.

        Args:
            hue (float): The hue component of the color, in the range [0, 360).
            saturation (float): The saturation component of the color,
                in the range [0, 1].
            value (float): The value (brightness) component of the color,
                in the range [0, 1].

        Returns:
            RGB: An RGB dataclass with red, green, and blue components.
        """
        return RGB(
            *tuple(
                round(i * 255, 0)
                for i in colorsys.hsv_to_rgb(hue / 360, saturation / 100, value / 100)
            )
        )

    @staticmethod
    def rgb_to_hsv(red: float, green: float, blue: float) -> HSV:
        """Convert RGB to HSV.

        Args:
            red (float): The red component of the color, in the range [0, 255].
            green (float): The green component of the color, in the range [0, 255].
            blue (float): The blue component of the color, in the range [0, 255].

        Returns:
            HSV: An HSV dataclass with hue, saturation, and value components.
        """
        hsv_tuple = colorsys.rgb_to_hsv(red / 255, green / 255, blue / 255)
        hsv_factors = [360, 100, 100]

        return HSV(
            float(round(hsv_tuple[0] * hsv_factors[0], 2)),
            float(round(hsv_tuple[1] * hsv_factors[1], 2)),
            float(round(hsv_tuple[2] * hsv_factors[2], 0)),
        )



"""
This tests all requests made by outlet devices.

All tests inherit from the TestBase class which contains the fixtures
and methods needed to run the tests.

The tests are automatically parametrized by `pytest_generate_tests` in
conftest.py. The two methods that are parametrized are `test_details`
and `test_methods`. The class variables are used to build the list of
devices, test methods and arguments.

The `helpers.call_api` method is patched to return a mock response.
The method, endpoint, headers and json arguments are recorded
in YAML files in the api directory, categorized in folders by
module and files by the class name.

The default is to record requests that do not exist and compare requests
that already exist. If the API changes, set the overwrite argument to True
in order to overwrite the existing YAML file with the new request.

See Also
--------
`utils.TestBase` - Base class for all tests, containing mock objects
`confest.pytest_generate_tests` - Parametrizes tests based on
    method names & class attributes
`call_json_outlets` - Contains API responses
"""

import pytest
import logging
import pyvesync.const as const
from pyvesync.base_devices.outlet_base import VeSyncOutlet
from base_test_cases import TestBase
from utils import assert_test, parse_args
import call_json
import call_json_outlets

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

OUTLET_DEV_TYPES = call_json_outlets.OUTLETS
POWER_METHODS = ['get_energy_update']
OUTLET_PARAMS = [[dev, method] for dev in OUTLET_DEV_TYPES for method in POWER_METHODS]


class TestOutlets(TestBase):
    """Outlets testing class.

    This class tests outlets device details and methods. The methods are
    parametrized from the class variables using `pytest_generate_tests`.
    The call_json_outlets module contains the responses for the API requests.
    The device is instantiated from the details provided by
    `call_json.DeviceList.device_list_item()`. Inherits from `utils.TestBase`.

    Instance Attributes
    -------------------
    self.manager : VeSync
        Instantiated VeSync object
    self.mock_api : Mock
        Mock with patched `helpers.call_api` method
    self.caplog : LogCaptureFixture
        Pytest fixture for capturing logs

    Class Attributes
    ---------------
    device : str
        Name of device type - outlets
    outlets : list
        List of device types for outlets, this variable is named
        after the device variable value
    base_methods : List[List[str, Dict[str, Any]]]
        List of common methods for all devices
    device_methods : Dict[List[List[str, Dict[str, Any]]]]
        Dictionary of methods specific to device types

    Methods
    --------
    test_details()
        Test the device details API request and response
    test_methods()
        Test device methods API request and response

    Examples
    --------
    >>> device = 'outlets'
    >>> outlets = ['ESW01-USA', 'ESW01-EU', 'ESW01-AU']
    >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
    >>> device_methods = {
        'dev_type': [['method1'], ['method2', {'kwargs': 'value'}]]
        }

    """

    device = 'outlets'
    outlets = call_json_outlets.OUTLETS
    base_methods = [
        ['turn_on'],
        ['turn_off']
    ]
    device_methods = {
        'ESW15-USA': [
            ['turn_on_nightlight'],
            ['turn_off_nightlight'],
            ['get_weekly_energy'],
            ['get_monthly_energy'],
            ['get_yearly_energy']
        ],
        'wifi-switch-1.3': [
            ['get_weekly_energy'],
            ['get_monthly_energy'],
            ['get_yearly_energy']
        ],
        # 'ESW03-USA': [
        #     ['get_weekly_energy'],
        #     ['get_monthly_energy'],
        #     ['get_yearly_energy']
        # ],
        'ESW01-EU': [
            ['get_weekly_energy'],
            ['get_monthly_energy'],
            ['get_yearly_energy']
        ],
        'ESO15-TB': [
            ['get_weekly_energy'],
            ['get_monthly_energy'],
            ['get_yearly_energy']
        ]
    }

    def test_details(self, setup_entry, method):
        """Test the device details API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class variables `device` (name of device type - outlets),
        device name (outlets) list of device types.

        Example:
            >>> device = 'outlets'
            >>> outlets = ['ESW01-USA', 'ESW01-EU', 'ESW01-AU']

        See Also
        --------
        `utils.TestBase` class docstring
        `call_json_outlets` module docstring

        Notes
        ------
        The device is instantiated using the `call_json.DeviceList.device_list_item()`
        method. The device details contain the default values set in `utils.Defaults`
        """        # Get response for device details
        details_response = call_json_outlets.DETAILS_RESPONSES[setup_entry]
        if callable(details_response):
            response_dict = details_response()
        else:
            response_dict = details_response
        self.mock_api.return_value = response_dict, 200
        # Get device configuration
        outlet_obj = self.get_device("outlets", setup_entry)
        assert isinstance(outlet_obj, VeSyncOutlet)

        # Call get_details() directly
        method_call = getattr(outlet_obj, method)
        self.run_in_loop(method_call)

        # Parse arguments from mock_api call into dictionary
        all_kwargs = parse_args(self.mock_api)

        # Set both write_api and overwrite to True to update YAML files
        assert assert_test(
            method_call,
            all_kwargs,
            setup_entry,
            write_api=self.write_api,
            overwrite=self.overwrite,
        )

        # Test bad responses
        self.mock_api.reset_mock()
        if setup_entry == 'wifi-switch-1.3':
            self.mock_api.return_value = (None, 400)
        else:
            self.mock_api.return_value = call_json.DETAILS_BADCODE, 200
        self.run_in_loop(outlet_obj.get_details)
        assert 'details' in self.caplog.text

    def test_methods(self, setup_entry: str, method):
        """Test device methods API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class variables `device` (name of device type - outlets),
        device name (outlets) list of device types, `base_methods` - list of
        methods for all devices, and `device_methods` - list of methods for
        each device type.

        Example:
            >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
            >>> device_methods = {
                'dev_type': [['method1'], ['method2', {'kwargs': 'value'}]]
                }

        Notes
        -----
        The response can be a callable that accepts the `kwargs` argument to
        sync the device response with the API response. In some cases the API
        returns data from the method call, such as `get_yearly_energy`, in other cases the
        API returns a simple confirmation the command was successful.

        See Also
        --------
        `TestBase` class method
        `call_json_outlets` module

        """        # Get method name and kwargs from method fixture
        method_name = method[0]
        if len(method) == 2 and isinstance(method[1], dict):
            method_kwargs = method[1]
        else:
            method_kwargs = {}

        # Set return value for call_api based on METHOD_RESPONSES
        method_response = call_json_outlets.METHOD_RESPONSES[setup_entry][method_name]
        if callable(method_response):
            if method_kwargs:
                resp_dict = method_response(**method_kwargs)
            else:
                resp_dict = method_response()
        else:
            resp_dict = method_response
        self.mock_api.return_value = resp_dict, 200
        # Get device configuration

        outlet_obj = self.get_device("outlets", setup_entry)
        assert isinstance(outlet_obj, VeSyncOutlet)

        # Get method from device object
        method_call = getattr(outlet_obj, method[0])

        # Ensure method runs based on device configuration
        if method[0] == 'turn_on':
            outlet_obj.state.device_status = const.DeviceStatus.OFF
        elif method[0] == 'turn_off':
            outlet_obj.state.device_status = const.DeviceStatus.ON

        # Call method with kwargs if present
        if method_kwargs:
            self.run_in_loop(method_call, **method_kwargs)
        else:
            self.run_in_loop(method_call)

        # Parse arguments from mock_api call into dictionary
        all_kwargs = parse_args(self.mock_api)

        # Assert request matches recorded request or write new records
        assert assert_test(
            method_call, all_kwargs, setup_entry, self.write_api, self.overwrite
        )

        # Test bad responses
        self.mock_api.reset_mock()
        if setup_entry == 'wifi-switch-1.3':
            self.mock_api.return_value = (None, 400)
        else:
            self.mock_api.return_value = call_json.DETAILS_BADCODE, 200
        if method[0] == 'turn_on':
            outlet_obj.state.device_status = const.DeviceStatus.OFF
        if method[0] == 'turn_off':
            outlet_obj.state.device_status = const.DeviceStatus.ON
        if 'energy' in method[0]:
            return
        bad_return = self.run_in_loop(method_call)
        assert bad_return is False

    @pytest.mark.parametrize('setup_entry', [d for d in OUTLET_DEV_TYPES])
    def test_power(self, setup_entry):
        """Test outlets power history methods."""
        outlet_obj = self.get_device("outlets", setup_entry)
        assert isinstance(outlet_obj, VeSyncOutlet)
        if not outlet_obj.supports_energy:
            pytest.skip(f"{setup_entry} does not support energy monitoring.")

        self.mock_api.side_effect = [
            (dict(call_json_outlets.METHOD_RESPONSES[setup_entry]['get_weekly_energy']), 200),
            (dict(call_json_outlets.METHOD_RESPONSES[setup_entry]['get_monthly_energy']), 200),
            (dict(call_json_outlets.METHOD_RESPONSES[setup_entry]['get_yearly_energy']), 200),

        ]

        self.run_in_loop(outlet_obj.update_energy)
        assert self.mock_api.call_count == 3
        assert outlet_obj.state.weekly_history is not None
        assert outlet_obj.state.monthly_history is not None
        assert outlet_obj.state.yearly_history is not None
        self.mock_api.reset_mock()
        if setup_entry == 'wifi-switch-1.3':
            self.mock_api.return_value = (None, 400)
        else:
            self.mock_api.return_value = call_json.DETAILS_BADCODE, 200



"""
This tests requests for Air Purifiers.

All tests inherit from the TestBase class which contains the fixtures
and methods needed to run the tests.

The tests are automatically parametrized by `pytest_generate_tests` in
conftest.py. The two methods that are parametrized are `test_details`
and `test_methods`. The class variables are used to build the list of
devices, test methods and arguments.

The `helpers.call_api` method is patched to return a mock response.
The method, endpoint, headers and json arguments are recorded
in YAML files in the api directory, categorized in folders by
module and files by the class name.

The default is to record requests that do not exist and compare requests
that already exist. If the API changes, set the overwrite argument to True
in order to overwrite the existing YAML file with the new request.

See Also
--------
`utils.TestBase` - Base class for all tests, containing mock objects
`confest.pytest_generate_tests` - Parametrizes tests based on
    method names & class attributes
`call_json_fans` - Contains API responses
"""

import logging
import pytest

import pyvesync.const as const
from pyvesync.base_devices.purifier_base import VeSyncPurifier
from base_test_cases import TestBase
from utils import assert_test, parse_args
import call_json_purifiers


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class TestAirPurifiers(TestBase):
    """Air Purifier testing class.

    This class tests Air Purifier device details and methods.
    The methods are parametrized from the class variables using
    `pytest_generate_tests`. The call_json_fans module contains the responses
    for the API requests. The device is instantiated from the details provided by
    `call_json.DeviceList.device_list_item()`. Inherits from `utils.TestBase`.

    Instance Attributes
    -------------------
    self.manager : VeSync
        Instantiated VeSync object
    self.mock_api : Mock
        Mock with patched `helpers.call_api` method
    self.caplog : LogCaptureFixture
        Pytest fixture for capturing logs

    Class Attributes
    ---------------
    device : str
        Name of product class - air_purifiers
    air_purifiers : list
        List of setup_entry's for air purifiers, this variable is named
        after the device variable value
    base_methods : List[List[str, Dict[str, Any]]]
        List of common methods for all devices
    device_methods : Dict[List[List[str, Dict[str, Any]]]]
        Dictionary of methods specific to device types

    Methods
    --------
    test_details()
        Test the device details API request and response
    test_methods()
        Test device methods API request and response

    Examples
    --------
    >>> device = 'air_purifiers'
    >>> air_purifiers = call_json_fans.AIR_MODELS
    >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
    >>> device_methods = {
        'ESWD16': [['method1'], ['method2', {'kwargs': 'value'}]]
        }
    """

    device = "air_purifiers"
    air_purifiers = call_json_purifiers.PURIFIER_MODELS
    base_methods = [
        ["turn_on"],
        ["turn_off"],
        ["set_sleep_mode"],
        ["set_manual_mode"],
        ["set_fan_speed", {"speed": 3}],
    ]
    # TODO: Add timer tests
    device_methods = {
        "Core200S": [["set_sleep_mode"],],
        "Core300S": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],
        "Core400S": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],
        "Core600S": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],
        "LV-PUR131S": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],
        "LV-RH131S": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],
        "LAP-V102S": [["set_auto_mode"], ['set_sleep_mode'], ['set_pet_mode'], ["turn_on_display"], ["turn_off_display"]],
        "LAP-V201S": [["set_auto_mode"], ['set_sleep_mode'], ['set_pet_mode'], ["turn_on_display"], ["turn_off_display"]],
        "EL551S": [["set_auto_mode"], ["set_sleep_mode"], ["set_turbo_mode"], ["turn_on_display"], ["turn_off_display"]],
        "LAP-B851S-WUS": [["set_auto_mode"], ['set_sleep_mode'], ["turn_on_display"], ["turn_off_display"]],

    }

    def test_details(self, setup_entry, method):
        """Test the device details API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class variables `device` (name of product class - air_purifiers),
        device name (air_purifiers) list of device types.

        Example:
            >>> device = 'air_purifiers'
            >>> air_purifiers = call_json_fans.AIR_MODELS

        See Also
        --------
        `utils.TestBase` class docstring
        `call_json_fans` module docstring

        Notes
        ------
        The device is instantiated using the `call_json.DeviceList.device_list_item()`
        method. The device details contain the default values set in `utils.Defaults`
        """
        # Set return value for call_api based on call_json_fan.DETAILS_RESPONSES
        return_dict = call_json_purifiers.DETAILS_RESPONSES[setup_entry]
        return_val = (return_dict, 200)
        self.mock_api.return_value = return_val

        # Instantiate device from device list return item
        fan_obj = self.get_device("air_purifiers", setup_entry)
        assert isinstance(fan_obj, VeSyncPurifier)

        method_call = getattr(fan_obj, method)
        self.run_in_loop(method_call)

        # Test state attributes
        assert fan_obj.state.device_status == const.DeviceStatus.ON
        assert fan_obj.state.mode == call_json_purifiers.PurifierDefaults.purifier_mode
        assert fan_obj.state.fan_level == call_json_purifiers.PurifierDefaults.fan_level

        # Parse mock_api args tuple from arg, kwargs to kwargs
        all_kwargs = parse_args(self.mock_api)

        # Assert request matches recorded request or write new records
        assert assert_test(
            method_call, all_kwargs, setup_entry, self.write_api, self.overwrite
        )

    def test_methods(self, setup_entry, method):
        """Test device methods API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class variables `device` (name of product class - air_purifiers),
        device name (air_purifiers) list of setup_entry's, `base_methods` - list of
        methods for all devices, and `device_methods` - list of methods for
        each device type.

        Example:
            >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
            >>> device_methods = {
                'setup_entry': [['method1'], ['method2', {'kwargs': 'value'}]]
                }

        Notes
        -----
        The response can be a callable that accepts the `kwargs` argument to
        sync the device response with the API response. In some cases the API
        returns data from the method call, such as `get_yearly_energy`, in other cases the
        API returns a simple confirmation the command was successful.

        See Also
        --------
        `TestBase` class method
        `call_json_fans` module

        """
        # TODO: FIX `clear_timer` recorded API request in yaml
        if method[0] == "clear_timer":
            pytest.skip("Incorrect clear_timer API request")
        # Get method name and kwargs from method fixture
        method_name = method[0]
        if len(method) == 2 and isinstance(method[1], dict):
            method_kwargs = method[1]
        else:
            method_kwargs = {}

        # Set return value for call_api based on call_json_fans.METHOD_RESPONSES
        method_response = call_json_purifiers.METHOD_RESPONSES[setup_entry][method_name]
        if callable(method_response):
            if method_kwargs:
                return_dict = method_response(method_kwargs)
            else:
                return_dict = method_response()
        else:
            return_dict = method_response
        self.mock_api.return_value = (return_dict, 200)

        # Get device configuration from call_json.DeviceList.device_list_item()
        fan_obj = self.get_device("air_purifiers", setup_entry)
        assert isinstance(fan_obj, VeSyncPurifier)

        # Get method from device object
        method_call = getattr(fan_obj, method_name)

        # Ensure method runs based on device configuration
        if method_name == "turn_on":
            fan_obj.state.device_status = const.DeviceStatus.OFF
        elif method_name == "turn_off":
            fan_obj.state.device_status = const.DeviceStatus.ON
        elif method_name == "set_fan_speed":
            fan_obj.state.mode = const.PurifierModes.MANUAL
        elif method_name in ["set_auto_mode", "set_sleep_mode", "set_pet_mode", "set_turbo_mode"]:
            fan_obj.state.mode = const.PurifierModes.MANUAL
        elif method_name == "set_manual_mode":
            fan_obj.state.mode = const.PurifierModes.AUTO
        elif method_name in ["turn_on_display"]:
            fan_obj.state.display_set_status = const.DeviceStatus.OFF
        elif method_name in ["turn_off_display"]:
            fan_obj.state.display_set_status = const.DeviceStatus.ON

        # Call method with kwargs if defined
        if method_kwargs:
            self.run_in_loop(method_call, **method_kwargs)
        else:
            self.run_in_loop(method_call)

        if method_name == "turn_on":
            assert fan_obj.state.device_status == const.DeviceStatus.ON
        elif method_name == "turn_off":
            assert fan_obj.state.device_status == const.DeviceStatus.OFF
        elif method_name == "set_fan_speed":
            assert fan_obj.state.fan_level == method_kwargs["speed"]
        elif method_name in ["set_auto_mode"]:
            assert fan_obj.state.mode == const.PurifierModes.AUTO
        elif method_name in ["set_sleep_mode"]:
            assert fan_obj.state.mode == const.PurifierModes.SLEEP
        elif method_name in ["set_pet_mode"]:
            assert fan_obj.state.mode == const.PurifierModes.PET
        elif method_name in ["set_turbo_mode"]:
            assert fan_obj.state.mode == const.PurifierModes.TURBO
        elif method_name in ["turn_on_display"]:
            assert fan_obj.state.display_set_status == const.DeviceStatus.ON
        elif method_name in ["turn_off_display"]:
            assert fan_obj.state.display_set_status == const.DeviceStatus.OFF

        # Parse arguments from mock_api call into a dictionary
        all_kwargs = parse_args(self.mock_api)

        # Assert request matches recorded request or write new records
        assert assert_test(
            method_call, all_kwargs, setup_entry, self.write_api, self.overwrite
        )



"""Base Device and State Class for VeSync Humidifiers."""

from __future__ import annotations

import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import DeviceStatus, DryingModes, HumidifierFeatures, HumidifierModes
from pyvesync.models.bypass_models import ResponseBaseModel

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import HumidifierMap
    from pyvesync.models.humidifier_models import SproutBreathinglamp
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class HumidifierState(DeviceState):
    """State Class for VeSync Humidifiers.

    This is the state class for all vesync humidifiers. If there are
    new features or state attributes.

    Attributes:
        auto_preference (int): Auto preference level.
        auto_stop_target_reached (bool): Automatic stop target reached.
        auto_target_humidity (int): Target humidity level.
        automatic_stop_config (bool): Automatic stop configuration.
        display_set_status (str): Display set status.
        display_status (str): Display status.
        drying_mode_auto_switch (str): Drying mode auto switch status.
        drying_mode_level (int): Drying mode level.
        drying_mode_status (str): Drying mode status.
        drying_mode_time_remain (int): Drying mode time remaining.
        filter_life_percent (int): Filter life percentage.
        humidity (int): Current humidity level.
        humidity_high (bool): High humidity status.
        mist_level (int): Mist level.
        mist_virtual_level (int): Mist virtual level.
        mode (str): Current mode.
        nightlight_brightness (int): Nightlight brightness level.
        nightlight_status (str): Nightlight status.
        temperature (float): Current temperature.
        warm_mist_enabled (bool): Warm mist enabled status.
        warm_mist_level (int): Warm mist level.
        water_lacks (bool): Water lacks status.
        water_tank_lifted (bool): Water tank lifted status.
    """

    __slots__ = (
        'auto_preference',
        'auto_stop_target_reached',
        'auto_target_humidity',
        'automatic_stop_config',
        'breathing_lamp',
        'child_lock',
        'display_set_status',
        'display_status',
        'drying_mode_auto_switch',
        'drying_mode_level',
        'drying_mode_status',
        'drying_mode_time_remain',
        'filter_life',
        'hepa_filter_life',
        'humidity',
        'humidity_high',
        'mist_level',
        'mist_virtual_level',
        'mode',
        'nightlight_brightness',
        'nightlight_color_temp',
        'nightlight_status',
        'temperature',
        'warm_mist_enabled',
        'warm_mist_level',
        'water_lacks',
        'water_lacks_drying_switch',
        'water_tank_lifted',
    )

    def __init__(
        self,
        device: VeSyncHumidifier,
        details: ResponseDeviceDetailsModel,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize VeSync Humidifier State.

        This state class is used to store the current state of the humidifier.

        Args:
            device (VeSyncHumidifier): The device object.
            details (ResponseDeviceDetailsModel): The device details.
            feature_map (HumidifierMap): The feature map for the device.
        """
        super().__init__(device, details, feature_map)
        self.auto_stop_target_reached: bool = False
        self.auto_target_humidity: int | None = None
        self.automatic_stop_config: bool = False
        self.display_set_status: str = DeviceStatus.UNKNOWN
        self.display_status: str = DeviceStatus.UNKNOWN
        self.humidity: int | None = None
        self.humidity_high: bool = False
        self.mist_level: int | None = None
        self.mist_virtual_level: int | None = None
        self.mode: str | None = None
        self.nightlight_brightness: int | None = None
        self.nightlight_status: str | None = None
        self.nightlight_color_temp: int | None = None
        self.warm_mist_enabled: bool | None = None
        self.warm_mist_level: int | None = None
        self.water_lacks: bool = False
        self.water_tank_lifted: bool = False
        self.child_lock: bool | None = None
        self.temperature: float | None = None  # Fahrenheit
        # Superior 6000S States / Sprout
        self.auto_preference: int | None = None
        self.filter_life: int | None = None
        self.drying_mode_level: int | None = None
        self.drying_mode_auto_switch: str | None = None
        self.drying_mode_status: DryingModes | None = None
        self.drying_mode_time_remain: int | None = None
        self.water_lacks_drying_switch: str | None = None
        self.hepa_filter_life: int | None = None
        self.breathing_lamp: BreathingLampState | None = None

    @property
    def automatic_stop(self) -> bool:
        """Return the automatic stop status.

        Returns:
            bool: True if automatic stop is enabled, False otherwise.
        """
        return self.automatic_stop_config

    @property
    def target_humidity(self) -> int | None:
        """Return the target humidity level.

        Returns:
            int: Target humidity level.
        """
        return self.auto_target_humidity

    @property
    def auto_humidity(self) -> int | None:
        """Return the auto humidity level.

        Returns:
            int: Auto humidity level.
        """
        return self.auto_target_humidity

    @property
    def auto_enabled(self) -> bool:
        """Return True if auto mode is enabled."""
        return self.mode in [HumidifierModes.AUTO, self.mode, HumidifierModes.HUMIDITY]

    @property
    def drying_mode_state(self) -> str | None:
        """Return the drying mode state.

        Returns:
            str | None: Drying mode state.
        """
        return self.drying_mode_status

    @property
    def drying_mode_seconds_remaining(self) -> int | None:
        """Return the drying mode seconds remaining.

        Return:
            int | None: Drying mode seconds remaining.
        """
        return self.drying_mode_time_remain

    @property
    def drying_mode_running(self) -> bool:
        """Return True if drying mode is enabled.

        Returns:
            bool | None: True if drying mode is enabled, False otherwise.
        """
        return self.drying_mode_status == DryingModes.RUNNING

    @property
    def breathing_lamp_status(self) -> str | None:
        """Return the breathing lamp status.

        Returns:
            str | None: Breathing lamp status.
        """
        if self.breathing_lamp is not None:
            return self.breathing_lamp.breathing_lamp_switch
        return None


class VeSyncHumidifier(VeSyncBaseToggleDevice):
    """VeSyncHumdifier Base Class.

    This is the base device to be inherited by all Humidifier devices.
    This class only holds the device configuration and static attributes.
    The state attribute holds the current state.

    Attributes:
        state (HumidifierState): The state of the humidifier.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        mist_levels (list): List of mist levels.
        mist_modes (list): List of mist modes.
        target_minmax (tuple): Tuple of target min and max values.
        warm_mist_levels (list): List of warm mist levels.

    """

    __slots__ = (
        '_reverse_mist_modes',
        'mist_levels',
        'mist_modes',
        'target_minmax',
        'warm_mist_levels',
    )

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: HumidifierMap,
    ) -> None:
        """Initialize VeSync Humidifier Class.

        Args:
            details (ResponseDeviceDetailsModel): The device details.
            manager (VeSync): The VeSync manager.
            feature_map (HumidifierMap): The feature map for the device.
        """
        super().__init__(details, manager, feature_map)
        self.state: HumidifierState = HumidifierState(self, details, feature_map)
        self.mist_modes: dict[str, str] = feature_map.mist_modes
        self._reverse_mist_modes: dict[str, str] = {}
        for k, v in self.mist_modes.items():
            self._reverse_mist_modes[v] = k
        self.mist_levels: list[int] = feature_map.mist_levels
        self.features: list[str] = feature_map.features
        self.warm_mist_levels: list[int] = feature_map.warm_mist_levels
        self.target_minmax: tuple[int, int] = feature_map.target_minmax

    @property
    def supports_warm_mist(self) -> bool:
        """Return True if the humidifier supports warm mist.

        Returns:
            bool: True if warm mist is supported, False otherwise.
        """
        return HumidifierFeatures.WARM_MIST in self.features

    @property
    def supports_nightlight(self) -> bool:
        """Return True if the humidifier supports nightlight.

        Returns:
            bool: True if nightlight is supported, False otherwise.
        """
        return HumidifierFeatures.NIGHTLIGHT in self.features

    @property
    def supports_nightlight_brightness(self) -> bool:
        """Return True if the humidifier supports nightlight brightness."""
        return HumidifierFeatures.NIGHTLIGHT_BRIGHTNESS in self.features

    @property
    def supports_drying_mode(self) -> bool:
        """Return True if the humidifier supports drying mode."""
        return HumidifierFeatures.DRYING_MODE in self.features

    async def toggle_automatic_stop(self, toggle: bool | None = None) -> bool:
        """Toggle automatic stop.

        Args:
            toggle (bool | None): True to enable automatic stop, False to disable.

        Returns:
            bool: Success of request.
        """
        del toggle
        if HumidifierFeatures.AUTO_STOP in self.features:
            logger.warning('Automatic stop has not been configured for this device.')
            return False
        logger.warning('Automatic stop is not supported or configured for this device.')
        return False

    async def toggle_display(self, toggle: bool) -> bool:
        """Toggle the display on/off.

        Args:
            toggle (bool): True to turn on the display, False to turn off.

        Returns:
            bool: Success of request.
        """
        del toggle
        logger.warning('Display is not supported or configured for this device.')
        return False

    @abstractmethod
    async def set_mode(self, mode: str) -> bool:
        """Set Humidifier Mode.

        Args:
            mode (str): Humidifier mode.

        Returns:
            bool: Success of request.

        Note:
            Modes for device are defined in `self.mist_modes`.
        """

    @abstractmethod
    async def set_mist_level(self, level: int) -> bool:
        """Set Mist Level for Humidifier.

        Args:
            level (int): Mist level.

        Returns:
            bool: Success of request.

        Note:
            Mist levels are defined in `self.mist_levels`.
        """

    async def turn_on_display(self) -> bool:
        """Turn on the display.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_display(True)

    async def turn_off_display(self) -> bool:
        """Turn off the display.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_display(False)

    async def turn_on_automatic_stop(self) -> bool:
        """Turn on automatic stop.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_automatic_stop(True)

    async def turn_off_automatic_stop(self) -> bool:
        """Turn off automatic stop.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_automatic_stop(False)

    async def set_auto_mode(self) -> bool:
        """Set Humidifier to Auto Mode.

        Returns:
            bool: Success of request.
        """
        if HumidifierModes.AUTO in self.mist_modes:
            return await self.set_mode(HumidifierModes.AUTO)
        logger.error('Auto mode not supported for this device.')
        return False

    async def set_manual_mode(self) -> bool:
        """Set Humidifier to Manual Mode.

        Returns:
            bool: Success of request.
        """
        if HumidifierModes.MANUAL in self.mist_modes:
            return await self.set_mode(HumidifierModes.MANUAL)
        logger.error('Manual mode not supported for this device.')
        return False

    async def set_sleep_mode(self) -> bool:
        """Set Humidifier to Sleep Mode.

        Returns:
            bool: Success of request.
        """
        if HumidifierModes.SLEEP in self.mist_modes:
            return await self.set_mode(HumidifierModes.SLEEP)
        logger.error('Sleep mode not supported for this device.')
        return False

    async def set_humidity(self, humidity: int) -> bool:
        """Set Humidifier Target Humidity.

        Args:
            humidity (int): Target humidity level.

        Returns:
            bool: Success of request.
        """
        del humidity
        logger.error('Target humidity is not supported or configured for this device.')
        return False

    async def set_nightlight_brightness(self, brightness: int) -> bool:
        """Set Humidifier night light brightness.

        Args:
            brightness (int): Target night light brightness.

        Returns:
            bool: Success of request.
        """
        del brightness
        if not self.supports_nightlight_brightness:
            logger.error('Nightlight brightness is not supported for this device.')
            return False
        logger.error('Nightlight brightness has not been configured.')
        return False

    async def toggle_nightlight(self, toggle: bool | None = None) -> bool:
        """Toggle night light on/off.

        Args:
            toggle (bool): True to turn on night light, False to turn off.

        Returns:
            bool: Success of request.
        """
        del toggle
        if not self.supports_nightlight:
            logger.error('Nightlight is not supported for this device.')
            return False
        logger.error('Nightlight has not been configured.')
        return False

    async def set_warm_level(self, warm_level: int) -> bool:
        """Set Humidifier Warm Level.

        Args:
            warm_level (int): Target warm level.

        Returns:
            bool: Success of request.
        """
        del warm_level
        if self.supports_warm_mist:
            logger.error('Warm level has not been configured.')
            return False
        logger.error('Warm level is not supported for this device.')
        return False

    async def turn_on_nightlight(self) -> bool:
        """Turn on night light.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_nightlight(True)

    async def turn_off_nightlight(self) -> bool:
        """Turn off night light.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_nightlight(False)

    async def toggle_drying_mode(self, toggle: bool | None = None) -> bool:
        """enable/disable drying filters after turning off."""
        del toggle
        if self.supports_drying_mode:
            logger.error('Drying mode is not configured for this device.')
            return False
        logger.error('Drying mode is not supported for this device.')
        return False

    async def turn_on_drying_mode(self) -> bool:
        """Turn on drying mode."""
        return await self.toggle_drying_mode(True)

    async def turn_off_drying_mode(self) -> bool:
        """Turn off drying mode."""
        return await self.toggle_drying_mode(False)

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        """Toggle child lock on/off.

        Args:
            toggle (bool): True to turn on child lock, False to turn off.

        Returns:
            bool: Success of request.
        """
        del toggle
        logger.error('Child lock is not supported or configured for this device.')
        return False

    async def turn_on_child_lock(self) -> bool:
        """Turn on child lock.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_child_lock(True)

    async def turn_off_child_lock(self) -> bool:
        """Turn off child lock.

        Returns:
            bool: Success of request.
        """
        return await self.toggle_child_lock(False)


@dataclass
class BreathingLampState(ResponseBaseModel):
    """Breathing Lamp State Object."""

    breathing_lamp_switch: str
    color_temperature: int
    time_interval: int
    brightness_start: int
    brightness_end: int

    @classmethod
    def from_model(cls, model: SproutBreathinglamp) -> BreathingLampState:
        """Create BreathingLampState from SproutBreathinglamp model.

        Args:
            model (SproutBreathinglamp): The SproutBreathinglamp model.

        Returns:
            BreathingLampState: The created BreathingLampState object.
        """
        return cls(
            breathing_lamp_switch=DeviceStatus.from_int(model.breathingLampSwitch),
            color_temperature=model.colorTemperature,
            time_interval=model.timeInterval,
            brightness_start=model.brightnessStart,
            brightness_end=model.brightnessEnd,
        )
